/*******************************************************************\

Module: Symbolic Execution of ANSI-C

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <fstream>
#include <iostream>
#include <memory>

#include <util/string2int.h>
#include <util/i2string.h>
#include <util/source_location.h>
#include <util/time_stopping.h>
#include <util/message_stream.h>

#include <langapi/mode.h>
#include <langapi/languages.h>
#include <langapi/language_util.h>

#include <ansi-c/ansi_c_language.h>

#include <goto-programs/xml_goto_trace.h>

#include <goto-symex/build_goto_trace.h>
#include <goto-symex/slice.h>
#include <goto-symex/slice_by_trace.h>
#include <goto-symex/memory_model_sc.h>
#include <goto-symex/memory_model_tso.h>
#include <goto-symex/memory_model_pso.h>

#include "bmc.h"
#include "bv_cbmc.h"
#include <algorithm>

/** added by sbjoshi **/
#include "../glue/glue_util.h"
#include "../glue/memory_model_tso_repair.h"
#include "../glue/memory_model_pso_repair.h"


/*******************************************************************\

Function: bmct::do_unwind_module

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::do_unwind_module(
  decision_proceduret &decision_procedure)
{
}

/*** added by sbjoshi ***/

/*******************************************************************\

Function : bmct::update_constraints

Input : target, prop_conv, ns, goto_trace, m

Output : 

Purpose : build the goto_trace and extract the context switch taken 

using these context switches compute the minimum hitting set m and bar 
the set of context switches given by m by adding assumptions to the solver


\*******************************************************************/
 void bmct::update_atomicity_constraints(symex_target_equationt &target,
  prop_conv_solvert &prop_conv,
  const namespacet &ns,
  goto_tracet &goto_trace,mhsit& m,atomicityt a)
 {
	 
	 
  build_goto_trace(equation, prop_conv, ns, goto_trace);

  //goto_trace.output(ns,std::cout);


  std::set<unsigned int> s;
  if(a==STRONG)
  {
  //std::list<concrete_event_structuret::concrete_eventst::const_iterator> sorted_events;
    // goto_trace.concrete_event_structure.get_sorted_events(ns,sorted_events,s);
  std::set<unsigned int> g_trace;
  get_cstaken_from_trace(goto_trace,s,g_trace);
  std::cout<<"CS Taken in the trace are " ;
  copy(s.begin(),s.end(),std::ostream_iterator<unsigned int>(std::cout,": "));
  std::cout<<endl;
	 if(s.size()==0)
	 {
		 std::cerr << "Every trace must take a context switch but the last trace does not...aborting" << std::endl;
		 assert(false);
	 }
	 satomic_repair.cpoints_collection.push_back(g_trace);
  }
  else
  {
    get_conflict_points_from_trace(goto_trace,s);

  }
	 m.add_to_collection(s);
  std::set<unsigned int> sol;
   m.get_mhs(sol);

   std::cout<< "Minimum hitting set now is : ";

   copy(sol.begin(),sol.end(),std::ostream_iterator<unsigned int>(std::cout,": "));


      add_atomicity_assumptions(prop_conv,sol,a);


 }

/****************************************************************************
 * Function : bmct::get_reordering_from_prop
 *
 * Input : prop_conv, s
 *
 * Output :
 *
 * Purpose : From the prop_conv, find out  which relaxed ordering took place.
 * One of these reordering has to be the culprit(from the assumption) thus
 * find out which literals (group of literals) needs to be set to TRUE.
 * A literal corresponding to the ordering, if TRUE will prevent the
 * reordering from happening
 *
 */
 void bmct::get_reordering_from_prop(prop_conv_solvert& prop_conv,
     std::set<unsigned int>& s, const literal2symbolt& literal2symbol,
     symbol2grpt& symbol2grp)
 {
    for(literal2symbolt::const_iterator it=literal2symbol.begin();
        it!=literal2symbol.end();it++)
    {
#ifdef DEBUG
    	std::cout << it->first.var_no() << ":" << prop_conv.l_get(it->first) << std::endl;
#endif
      if(prop_conv.l_get(it->first)!=tvt(true)) continue;

      irep_idt tmp = it->second;
#ifdef DEBUG
       std::cout << "group " << symbol2grp[tmp] << " inserted" << std::endl;
       std::cout << symbol2order[tmp].first->source.pc->source_location.as_string() << " and " << \
    		   symbol2order[tmp].second->source.pc->source_location.as_string() << " reordered" << std::endl;
#endif
      s.insert(symbol2grp[tmp]);
    }
 }
/*****************************************************************************\
 * Function : bmct::add_reordering_assumptions
 *
 * Input : prop_conv, ban_reord
 *
 * Output : ban_reord represents the group of literals which should be set
 * to TRUE to prevent culprit reorderings. assumptions are added to the
 * formula (prop_conv) to reflect the same.
 *
 * Purpose :
 *
 */
 void bmct::add_reordering_assumptions(prop_conv_solvert& prop_conv,
     std::set<unsigned int>& ban_reord, grp2grplitt& grp2grplit, bvt& prev_assumptions) const
 {
   bvt assumptions;
  // ban_reord.insert(0);ban_reord.insert(1);ban_reord.insert(2);ban_reord.insert(3);ban_reord.insert(4);ban_reord.insert(5);


   for(std::set<unsigned int>::iterator it=ban_reord.begin();
       it!=ban_reord.end();it++)
   {

    unsigned int g=(*it);
    assumptions.push_back(!grp2grplit[g]);
/*
   // orderingliteralst::const_iterator oit= oliterals.find(g);

     for(literalst::const_iterator lit= oliterals[g].begin();
         lit!=oliterals[g].end();lit++)
     {
    	 //set the literal to false to make the corresponding boolean var to
    	 //true
       literalt l(lit->var_no(),true);
       assumptions.push_back(l);
     }*/
   }
   assumptions.insert(assumptions.begin(),prev_assumptions.begin(),prev_assumptions.end());
   prop_conv.set_assumptions(assumptions);
 }

 void bmct::add_reordering_memorax_constraint(propt& prop,
		 std::set<unsigned int>& gtaken, grp2grplitt& grp2grplit) const
 {

	 bvt constraint;
	 for(std::set<unsigned int>::iterator it=gtaken.begin();it!=gtaken.end();
			 it++)
	 {
		 constraint.push_back(!grp2grplit[*it]);
	 }
	 prop.lcnf(constraint);
 }

 /****************************************************************************\
  *
  * Function : bmct::update_ordering_constraints
  *
  * Input : prop_conv, m
  *
  * Output :
  *
  * Purpose : Given the satisfiable evaluation in prop_conv (due to assertion
  * violation), find out the least (through Minimum-hitting set in m) number of
  * reorderings to be prevented in order to avoid assertion violation.
  */
 void bmct::update_ordering_constraints(
   prop_conv_solvert &prop_conv, mhsit& m,literal2symbolt& literal2symbol,
   symbol2grpt& symbol2grp,grp2grplitt& grp2grplit, bvt& prev_assumptions)
  {





   std::set<unsigned int> s;

   get_reordering_from_prop(prop_conv,s,literal2symbol,symbol2grp);

    if(s.size()==0)
    {
      std::cerr << "Every trace must have at least one re-ordering. " \
          "Found none...aborting" << std::endl;
      assert(false);
    }
#ifdef DEBUG
    std::cout << "groups taken ";
    std::copy(s.begin(),s.end(),std::ostream_iterator<unsigned int>(std::cout," "));
    std::cout << std::endl;
#endif
    m.add_to_collection(s);
   std::set<unsigned int> sol;
   //get MHS-Minimum Hitting Set in sol
   m.get_mhs(sol);
#ifdef DEBUG
    std::cout<< "Minimum hitting set now is : ";

    copy(sol.begin(),sol.end(),
        std::ostream_iterator<unsigned int>(std::cout,": "));
#endif


       add_reordering_assumptions(prop_conv,sol,grp2grplit,prev_assumptions);

  }
 void bmct::update_ordering_constraints(
      prop_conv_solvert &prop_conv, propt& prop, mhsit& m,literal2symbolt& literal2symbol,
      symbol2grpt& symbol2grp,grp2grplitt& grp2grplit)
 {





   std::set<unsigned int> s;

   get_reordering_from_prop(prop_conv,s,literal2symbol,symbol2grp);

    if(s.size()==0)
    {
      std::cerr << "Every trace must have at least one re-ordering. " \
          "Found none...aborting" << std::endl;
      assert(false);
    }
#ifdef DEBUG
    std::cout << "groups taken ";
    std::copy(s.begin(),s.end(),std::ostream_iterator<unsigned int>(std::cout," "));
    std::cout << std::endl;
#endif
    m.add_to_collection(s);

add_reordering_memorax_constraint(prop,s,grp2grplit);



  }

 /***************************************************************************\
 Function : bmct::add_atomicity_assumptions

 Input : prop_conv, ban_cs

 Purpose : add the assumptions to prop so as to ban the context switches
 denoted by ban_cs
 \***************************************************************************/
 void bmct::add_atomicity_assumptions(prop_conv_solvert& prop_conv,
     std::set<unsigned int>& ban_cs,atomicityt a) const
 { bvt assumptions;

//ban_cs.insert(7);
//ban_cs.insert(9);
   const guard_to_literal_mapt& gmap = a==STRONG? guard_to_literal_map: wbguard_to_literal_map;
	 for(std::set<unsigned int>::const_iterator it=ban_cs.begin();it!=ban_cs.end();it++)
	 {
		 unsigned int guard = (*it);
		 //disable all cs for debugging -- remove later
	 //for(guard_to_literal_mapt::const_iterator git=gmap.begin();
//			 git!=gmap.end();git++)
	// {

		 std::cout << "Protecting guard : " << guard << std::endl;
		 for(literalst::const_iterator lit = gmap[guard-1].begin();
			 lit!=gmap[guard-1].end();lit++)

		 //for(literalst::const_iterator lit=git->begin();lit!=git->end();
		//		 lit++)
		 {
		   bool val = a==STRONG?false:true;

		//   std::cout << (*lit).var_no() << " : " << val;
			 literalt l((*lit).var_no(),val);
			 // making literal positive merges two consecutive atomic sections (for strong)

			 // making literal negative so that conflict points reach outside the
			 //locked region (for weak)
			// std::cout << "var : " << (*lit).var_no() << "being positively taken" << std::endl;
			 assumptions.push_back(l);

		 }
		// }
	}

	 prop_conv.set_assumptions(assumptions);
 }

/*****************************************************************\

Function : print_guard_to_literal_map

Input : std::ostream &out

Output : 

Purpose : Prints the guard_to_literal_map onto a give ostream "out"
\*****************************************************************/

 void bmct::print_guard_to_literal_map(std::ostream &out,atomicityt a)
 {
   guard_to_literal_mapt& g_map = (a==STRONG)?guard_to_literal_map:wbguard_to_literal_map;
	 for(guard_to_literal_mapt::const_iterator vit = g_map.begin();
		 vit!=g_map.end();vit++)
	 {
		 out << (vit-g_map.begin())+1 << " : ";
		 for(literalst::const_iterator lit=vit->begin();
			 lit!=vit->end();lit++)
		 {
			 out << lit->dimacs() << " ";
		 }
		 out << std::endl;
	 }
 }
 /*****************************************************************\

 Function : print_guard_to_literal_map

 Input : std::ostream &out

 Output :

 Purpose : Prints the guard_to_literal_map onto a give ostream "out"
 \*****************************************************************/
#if 0
  void bmct::print_wbguard_to_literal_map(std::ostream &out)
  {
    for(guard_to_literal_mapt::const_iterator vit = guard_to_literal_map.begin();
      vit!=guard_to_literal_map.end();vit++)
    {
      for(literalst::const_iterator lit=vit->begin();
        lit!=vit->end();lit++)
      {
        out << lit->dimacs() << " ";
      }
      out << std::endl;
    }
  }
#endif
/******************************************************************\
Function : initialize_guard_literal_map

Input : size

Output : 

Purpose : Initialize the guard_to_literal_map with required number of
elements and pushes an empty list in each of them.
\******************************************************************/

void bmct::initialize_guard_literal_map(unsigned int size,atomicityt a)
{
  guard_to_literal_mapt& g_map = (a==STRONG)?guard_to_literal_map:wbguard_to_literal_map;

	g_map.reserve(size);

	for(unsigned int i=0;i<size;i++)
	{
		std::list<literalt> l;
		g_map.push_back(l);

	}
}
#if 0
/******************************************************************\
Function : initialize_guard_literal_map

Input : size

Output :

Purpose : Initialize the guard_to_literal_map with required number of
elements and pushes an empty list in each of them.
\******************************************************************/

void bmct::initialize_wbguard_literal_map(unsigned int size)
{
  wbguard_to_literal_map.reserve(size);

  for(unsigned int i=0;i<size;i++)
  {
    std::list<literalt> l;
    wbguard_to_literal_map.push_back(l);

  }
}
#endif

/****************************************************************************\
 * Function : pupulate_oliterals_map
 *
 * Input : prop_convt
 *
 * Output : None
 *
 * Purpose : In symex equation, constraints are added through non-det
 * auxiliary boolean symbols. If the symbols is true then the program order
 * is NOT relaxed. If it is false then the program order is relaxed. Multiple
 * event pair may be instances of the same pair of instructions/variables. When
 * we want to restrict the ordering, we want to restrict the ordering of all
 * instances.
 *
 * Given an irep_idt (non det boolean symbol identifier), symbol2grp
 * maps it to a group number. All symbols belonging to the same group represents
 * different instance of the same event pairs. From prop_conv, we find out the
 * literals of symbols and insert it in oliterals in its corresponding group.
 *
 * The reason why we store literal2symbol mapping is because from prop
 * we can query through literals only. So from the literals we find
 * the symbols and then find the group of literals that restricts the
 * ordering.
 *
 * The reason why symbols are still need to be carried around is that
 * the event pair information associated with them is needed to show
 * results (for user interface) as to what events(lines/instructions)
 * need to be protected from being reordered.
 */
void bmct::populate_oliterals_map(prop_conv_solvert& prop_conv, literal2symbolt& literal2symbol,
		symbol2grpt& symbol2grp, orderingliteralst& oliterals,grp2grplitt& grp2grplit)
{
	grp2grplit.clear();
	  std::map<unsigned int,unsigned int> oldg2newg;
  for(prop_conv_solvert::symbolst::const_iterator s_it=prop_conv.get_symbols().begin();
        s_it!=prop_conv.get_symbols().end();
        s_it++)
  {


    symbol2grpt::iterator it= symbol2grp.find(s_it->first);
    if(it!=symbol2grp.end())
    {
    	if(does_irep_id_contains_str(s_it->first,GLUE_GROUP_SYMBOL))
    		continue;
    	if(oldg2newg.find(it->second)==oldg2newg.end())
    	{
    		literalst l;
    		//oliterals[it->second]=l;
    		oliterals.push_back(l);
    		oldg2newg[it->second] = oliterals.size()-1;
    	}
       oliterals[oldg2newg[it->second]].push_back(s_it->second);
       literal2symbol[s_it->second]=it->first;


#ifdef DEBUG
      std::cout << it->first << ":" << it->second << " : "<< s_it->second.var_no() << ":" << \
    		  symbol2order[it->first].first->source.pc->location_number << ":" << \
    		  symbol2order[it->first].first->source.pc->source_location.as_string() << \
    		  symbol2order[it->first].second->source.pc->location_number << ":" << \
    		  symbol2order[it->first].second->source.pc->source_location.as_string() << std::endl;
#endif


      prop_conv.set_frozen(literalt(s_it->second));
    }


  }

  unsigned int i=0;
  for(orderingliteralst::iterator oit = oliterals.begin();
		  oit!=oliterals.end();oit++)
  {
	  std::cout << "group " << i << " : ";
	  for(literalst::iterator lit=oliterals[i].begin();lit!=oliterals[i].end();lit++)
	  {
		  symbol2grp[literal2symbol[(*lit)]]=i;
#ifdef DEBUG
		  std::cout << lit->var_no() << " ";
#endif
	  }
#ifdef DEBUG
	  std::cout << std::endl;
#endif
	  //initialize with random literal
grp2grplit.push_back(literalt());
	  i++;
  }

  //populate map from final group no to group literal
  for(prop_conv_solvert::symbolst::const_iterator s_it=prop_conv.get_symbols().begin();
          s_it!=prop_conv.get_symbols().end();
          s_it++)
  {
	  if(does_irep_id_contains_str(s_it->first,GLUE_GROUP_SYMBOL))
	  {
		  assert(!s_it->second.is_constant());
		  //a given group symbol MUST appear only one otherwise this fails
		  symbol2grp[s_it->first] = oldg2newg[symbol2grp[s_it->first]];
#ifdef DEBUG
		  std::cout << "group" << symbol2grp[s_it->first] << ": " << s_it->second << std::endl;
#endif
		  grp2grplit[symbol2grp[s_it->first]]=s_it->second;
		  prop_conv.set_frozen(literalt(s_it->second));
	  }
  }

}
/******************************************************************\
Function: populate_guard_literal_map

Input: prop_convt

Output : 

Purpose : After do_conversion is invoked, prop_convt::symbols map 
is populated with symbol->literal mapping. bmc::guard_to_literal_map
needs to be populated from such that literal corresponding to say
glue_bguard_5 can be accessed through guard_to_literal_map[4]

Precondition : This function relies on the fact that all the
original variables $glue_bguard_counter_XX is renamed as
YY$glue_bguard_counter_XX#ZZ where YY is any string, XX is the number
of the guard and ZZ could again be any string.

The way it extracts number is by finding the position at which $glue_bguard..
is found and where # is found. Then anything that is after $glue_bguard_counter_
and before # MUST be the number of the guard.
\******************************************************************/

void bmct::populate_guard_literal_map(const prop_conv_solvert& prop_conv,atomicityt a)
{

guard_to_literal_mapt& g_map = (a==STRONG)? guard_to_literal_map:wbguard_to_literal_map;
/*if(a==STRONG)
{
  g_map = guard_to_literal_map;
}
else
{
  g_map = wbguard_to_literal_map;
}*/
	std::string str1 = a==STRONG?GLUE_BGUARD:GLUE_WBGUARD;
	std::size_t len = str1.size();
	{
	unsigned int counter = a==STRONG?satomic_repair.get_bguard_counter():watomic_repair.get_wbguard_counter();
	initialize_guard_literal_map(counter,a);
	}

	//std::cout << "INSIDE POPULATE GUARD LITERAL" << std::endl;
   for(prop_conv_solvert::symbolst::const_iterator
      s_it=prop_conv.get_symbols().begin();
      s_it!=prop_conv.get_symbols().end();
      s_it++)
  {
    

		//std::cout << s_it->first << " : " << s_it->second.dimacs() << std::endl;
		std::string s1(as_string(s_it->first));
		std::size_t pos1 = s1.find(str1);
		if(pos1 != std::string::npos)
		{
			std::size_t pos2 = s1.find("#");
			std::string numstr;
			if (pos2 != std::string::npos)
			{
				numstr = s1.substr(pos1+len,pos2-(pos1+len));
			}
			else
			{
			  numstr=s1.substr(pos1+len);
			}
				std::istringstream is(numstr);
				    unsigned int guard_no;
				    is >> guard_no;

				    if((is.rdstate() & std::istringstream::failbit)
				        ||( is.rdstate() & (std::istringstream::badbit)) )
				       {
                  //no valid number could be found!!
				           assert(false);
				       }

				assert(guard_no>0);

				g_map[guard_no-1].push_back(s_it->second);

				if(s_it->second.is_constant()) {assert(false);}
			//	std::cout << s_it->first << " : " << s_it->second.dimacs() << " : " << guard_no << std::endl;
		/*	}

			else
			{
				//we have encountered an ill-formed string. The guardname should be of the kind 
				// c::$glue_bguard_XXX#YY where XXX and YY are numbers. We are interested in XXX part

				assert(false);
			}*/
			



		}
	
  }

	  print_guard_to_literal_map(std::cout,a);

}
/****************************************************************************\
 * Function : group_from_ordering
 *
 * Input :
 *
 * Output :
 *
 *
 * Purpose : symbol2order is populated via memory_model_xxx_repairt.
 * symbol2order is a map from non-det auxiliary boolean symbol to event pairs.
 * If symbol is true then event pairs observe the program order otherwise
 * the program order is relaxed. Since multiple event pairs can be generated
 * from same instruction/variable (through unwinding), we want to group them
 * together. All event pairs belonging to the same instructions/variable pair
 * will have the same group id in symbol2grp map.
 *
 * This function also initializes the oliterals map (group id to list of
 * literals). This will enable to restrict reordering of all the instances
 * of a particular pair in one go.
 */
void bmct::group_from_ordering(symbol2grpt& symbol2grp, sym2grpsymt& sym2grpsym)
{
  unsigned int group_no=0;

  typedef std::vector<std::list<irep_idt> > collectiont;
  collectiont collection;


     symbol2ordert::const_iterator sit_end = symbol2order.end();
     for(symbol2ordert::const_iterator sit=symbol2order.begin();
         sit!=sit_end;sit++)
     {
       symbol2ordert::const_iterator sit1 = sit;
       ++sit1;
        for(;sit1!=sit_end;sit1++)
        {
          unsigned int g1,g2;
          bool flag1=false,flag2=false;

          //find if group id is assigned to either sit or sit1
          if(symbol2grp.find(sit->first)!=symbol2grp.end())
          {
            flag1=true;g1=symbol2grp[sit->first];
          }
          if(symbol2grp.find(sit1->first)!=symbol2grp.end())
          {
            flag2=true;g2=symbol2grp[sit1->first];
          }
          //assuming that the instructions are preserved even after unwinding
          //continue if the events belong to different instructions or
          // same instructions but different shared variables
          peventt p = sit->second, p1=sit1->second;
          if((p.first->source.pc != p1.first->source.pc)

            || (p.second->source.pc != p1.second->source.pc)
            || (address_of(p.first) != address_of(p1.first))
            || (address_of(p.second) != address_of(p1.second)))
          {
            // if group id is not assigned then assign one
            // for the symbol
            if(!flag1)
            {
              symbol2grp[sit->first]=group_no;
              std::list<irep_idt> s;
              s.push_back(sit->first);
              collection.push_back(s);
              group_no++;
            }
            if(!flag2)
            {
              symbol2grp[sit1->first] = group_no;
              std::list<irep_idt> s;
              s.push_back(sit1->first);
              collection.push_back(s);
              group_no++;
            }

          }
          //if events comes from the same variables group them together
          else
          {
            //first two cases represents when only one of the symbol
            // has a valid group id. Just add another with the first one
             if(flag1 && !flag2)
             {

               symbol2grp[sit1->first] = symbol2grp[sit->first];
               collection[g1].push_back(sit1->first);
             }
             else if(flag2 && !flag1)
             {
               symbol2grp[sit->first] = symbol2grp[sit1->first];
               collection[g2].push_back(sit->first);
             }

             //when both of the symbols do not have group ids
             else if(!flag1 && !flag2)
             {
               symbol2grp[sit->first] = symbol2grp[sit1->first]=group_no;
               std::list<irep_idt> s;
               s.push_back(sit->first); s.push_back(sit1->first);
               collection.push_back(s);
               group_no++;
             }

             // when both symbols have valid group id but are different
             // in this case the group has to be merged
             else if(g1!=g2)
             {
               // always merge the larger id group with the smaller id

               if(g1 > g2)
               {
            	   // swap g1 with g2 to make g1 smaller
                 unsigned int tmp = g1;
                 g1=g2;
                 g2=tmp;
               }
               for(std::list<irep_idt>::iterator it=collection[g2].begin(),it_end=collection[g2].end();
                   it!=it_end;it++)
               {
                 symbol2grp[(*it)] = g1;
                 collection[g1].push_back(*it);
               }
               collection[g2].clear();


             }
             //the final else case is when both symbols have valid grp ids
             // and they are the same, in which case nothing needs to be done

          }



        }

     }
//std::cout << "Collection size initially " << collection.size() << std::endl;
     collectiont::iterator cit=collection.begin();
     while(cit!=collection.end())
     {
       if(cit->empty())
       {
    	   collection.erase(cit);
       }
       if(cit==collection.end()) break;
    	 cit++;
     }
  //   std::cout << "Collection size after " << collection.size() << std::endl;

     unsigned csize=collection.size();
     for(int i=0;i<csize;i++)
     {
    	 symbol_exprt gsymbol(GLUE_GROUP_SYMBOL+i2string(i),bool_typet());
    	 for(typename std::list<irep_idt>::iterator lit=collection[i].begin();
    			 lit!=collection[i].end();lit++)
    	 {

    		 symbol_exprt order_symbol((*lit),bool_typet());
    		 equation.constraint(implies_exprt(not_exprt(gsymbol),
    				 not_exprt(order_symbol)),
    				 "glue_group_"+i2string(i),
    				 symbol2order[(*lit)].first->source);
    		 symbol2grp[(*lit)] = i;
    		 sym2grpsym[(*lit)]=gsymbol.get_identifier();
    	 }


    	 symbol2grp[gsymbol.get_identifier()]=i;
    	 //symbol2order[gsymbol.get_identifier()] = symbol2order[collection[i].front()];
     }
/*     //initialize oliterals here. Due to merging, max may be smaller than group_no
     unsigned int max=0;
     for(symbol2grpt::iterator it=symbol2grp.begin();it!=symbol2grp.end();it++)
     {
        if(it->second > max) max = it->second;
     }

     // max represents the highest group id, since group ids start from 0
     //the size should be max+1
      *
      */
     // add empty list to the vectors so that we need not worry later if
     // a particular entry contains a valid list or not.
//     oliterals.resize(csize);

#if 0
     for(unsigned int i=0;i<csize;i++)
     {
       literalst ll;
       oliterals.push_back(ll);
     }
#endif
}


/******************************************************************\
Function: populate_guard_literal_map

Input: prop_convt

Output :

Purpose : After do_conversion is invoked, prop_convt::symbols map
is populated with symbol->literal mapping. bmc::guard_to_literal_map
needs to be populated from such that literal corresponding to say
glue_bguard_5 can be accessed through guard_to_literal_map[4]

Precondition : This function relies on the fact that all the
original variables $glue_bguard_counter_XX is renamed as
YY$glue_bguard_counter_XX#ZZ where YY is any string, XX is the number
of the guard and ZZ could again be any string.

The way it extracts number is by finding the position at which $glue_bguard..
is found and where # is found. Then anything that is after $glue_bguard_counter_
and before # MUST be the number of the guard.
\******************************************************************/
#if 0
void bmct::populate_wbguard_literal_map(prop_convt& prop_conv)
{


  std::string str1(GLUE_WBGUARD);
  std::size_t len = str1.size();

  initialize_guard_literal_map(satomic_repair.get_bguard_counter());

  std::cout << "INSIDE POPULATE GUARD LITERAL" << std::endl;
   for(prop_convt::symbolst::const_iterator
      s_it=prop_conv.get_symbols().begin();
      s_it!=prop_conv.get_symbols().end();
      s_it++)
  {


    std::cout << s_it->first << " : " << s_it->second.dimacs() << std::endl;
    std::string s1(as_string(s_it->first));
    std::size_t pos1 = s1.find(str1);
    if(pos1 != std::string::npos)
    {
      std::size_t pos2 = s1.find("#");

      if (pos2 != std::string::npos)
      {
        std::string numstr = s1.substr(pos1+len,pos2-(pos1+len));
        std::istringstream is(numstr);
            unsigned int guard_no;
            is >> guard_no;

            if((is.rdstate() & std::istringstream::failbit)
                ||( is.rdstate() & (std::istringstream::badbit)) )
               {
                  //no valid number could be found!!
                   assert(false);
               }

        assert(guard_no>0);

        wbguard_to_literal_map[guard_no-1].push_back(s_it->second);

        if(s_it->second.is_constant()) {assert(false);}
        std::cout << s_it->first << " : " << s_it->second.dimacs() << " : " << guard_no << std::endl;
      }
      else
      {
        //we have encountered an ill-formed string. The guardname should be of the kind
        // c::$glue_bguard_XXX#YY where XXX and YY are numbers. We are interested in XXX part

        assert(false);
      }




    }

  }

    print_wbguard_to_literal_map(std::cout);

}


#endif
/*******************************************************************\

Function: bmct::error_trace

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::error_trace(const prop_convt &prop_conv)
{
  status() << "Building error trace" << eom;

  goto_tracet goto_trace;
  build_goto_trace(equation, prop_conv, ns, goto_trace);
  
  #if 0
  if(options.get_option("vcd")!="")
  {
    if(options.get_option("vcd")=="-")
      output_vcd(ns, goto_trace, std::cout);
    else
    {
      std::ofstream out(options.get_option("vcd").c_str());
      output_vcd(ns, goto_trace, out);
    }
  }
  #endif
  
  switch(ui)
  {
  case ui_message_handlert::PLAIN:
    std::cout << "\n" << "Counterexample:" << "\n";
    show_goto_trace(std::cout, ns, goto_trace);
    break;
  
  case ui_message_handlert::XML_UI:
    {
      xmlt xml;
      convert(ns, goto_trace, xml);
      std::cout << xml << "\n";
    }
    break;
  
  default:
    assert(false);
  }
}

/*******************************************************************\

Function: bmct::do_conversion

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::do_conversion(prop_convt &prop_conv)
{
  // convert HDL
  do_unwind_module(prop_conv);

  // convert SSA
  equation.convert(prop_conv);

  // the 'extra constraints'
  forall_expr_list(it, bmc_constraints)
    prop_conv.set_to_true(*it);
}

/*******************************************************************\

Function: bmct::run_decision_procedure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

decision_proceduret::resultt
bmct::run_decision_procedure(prop_convt &prop_conv)
{
  status() << "Passing problem to " 
           << prop_conv.decision_procedure_text() << eom;

  prop_conv.set_message_handler(get_message_handler());

  // stop the time
  absolute_timet sat_start=current_time();
  
  do_conversion(prop_conv);  

  status() << "Running " << prop_conv.decision_procedure_text() << eom;

  decision_proceduret::resultt dec_result=prop_conv.dec_solve();
  // output runtime

  {
    absolute_timet sat_stop=current_time();
    status() << "Runtime decision procedure: "
             << (sat_stop-sat_start) << "s" << eom;
  }

  return dec_result;
}

/*******************************************************************\

Function: bmct::report_success

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::report_success()
{
  result() << "VERIFICATION SUCCESSFUL" << eom;

  switch(ui)
  {
  case ui_message_handlert::PLAIN:
    break;
    
  case ui_message_handlert::XML_UI:
    {
      xmlt xml("cprover-status");
      xml.data="SUCCESS";
      std::cout << xml;
      std::cout << "\n";
    }
    break;
    
  default:
    assert(false);
  }
}

/*******************************************************************\

Function: bmct::report_failure

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::report_failure()
{
  result() << "VERIFICATION FAILED" << eom;

  switch(ui)
  {
  case ui_message_handlert::PLAIN:
    break;
    
  case ui_message_handlert::XML_UI:
    {
      xmlt xml("cprover-status");
      xml.data="FAILURE";
      std::cout << xml;
      std::cout << "\n";
    }
    break;
    
  default:
    assert(false);
  }
}

/*******************************************************************\

Function: bmct::show_program

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::show_program()
{
  unsigned count=1;

  languagest languages(ns, new_ansi_c_language());
  
  std::cout << "\n" << "Program constraints:" << "\n";

  for(symex_target_equationt::SSA_stepst::const_iterator
      it=equation.SSA_steps.begin();
      it!=equation.SSA_steps.end(); it++)
  {
    if(it->is_assignment())
    {
      std::string string_value;
      languages.from_expr(it->cond_expr, string_value);
      std::cout << "(" << count << ") " << string_value << "\n";

      if(!it->guard.is_true())
      {
        languages.from_expr(it->guard, string_value);
        std::cout << std::string(i2string(count).size()+3, ' ');
        std::cout << "guard: " << string_value << "\n";
      }
      
      count++;
    }
    else if(it->is_assert())
    {
      std::string string_value;
      languages.from_expr(it->cond_expr, string_value);
      std::cout << "(" << count << ") ASSERT("
                << string_value <<") " << "\n";

      if(!it->guard.is_true())
      {
        languages.from_expr(it->guard, string_value);
        std::cout << std::string(i2string(count).size()+3, ' ');
        std::cout << "guard: " << string_value << "\n";
      }

      count++;
    }  
  }
}

/*******************************************************************\

Function: bmct::run

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool bmct::run(const goto_functionst &goto_functions)
{
  const std::string mm=options.get_option("mm");
  unsigned int tmp = options.get_unsigned_int_option("repair");
  repairmodet rm;
  if(tmp)
  {
   rm =  static_cast<repairmodet>(tmp);
  }

  std::auto_ptr<memory_model_baset> memory_model(0);
  if(rm==SATOMIC || rm==WATOMIC || mm=="sc" || (mm.empty()&& tmp==0))
    memory_model.reset(new memory_model_sct(ns));
  else if(mm=="tso")
    memory_model.reset(new memory_model_tsot(ns));
  else if(mm=="pso")
    memory_model.reset(new memory_model_psot(ns));
  else if(rm==TSOSC || rm==PSOTSO || rm==PSOSC)
  {

     if(rm==TSOSC)
     {
       memory_model.reset(new memory_model_tso_repairt(symbol2order,ns));
     }
     else if(rm==PSOTSO)
     {
       memory_model.reset(new memory_model_pso_repairt(symbol2order,ns,memory_model_pso_repairt::TSO));
     }
     else if(rm==PSOSC)
     {
       memory_model.reset(new memory_model_pso_repairt(symbol2order,ns));
     }

  }
  else
    throw "Invalid memory model "+mm+" -- use one of sc, tso, pso";

  //symex.total_claims=0;
  symex.set_message_handler(get_message_handler());
  symex.options=options;

  status() << "Starting Bounded Model Checking" << eom;

  symex.last_source_location.make_nil();

  try
  {
    // get unwinding info
    setup_unwind();

    // perform symbolic execution
    symex(goto_functions);

    // add a partial ordering, if required    
    if(equation.has_threads())
    {
      memory_model->set_message_handler(get_message_handler());
      (*memory_model)(equation);
    }
  }

  catch(std::string &error_str)
  {
    message_streamt message_stream(get_message_handler());
    message_stream.err_location(symex.last_source_location);
    message_stream.error(error_str);
    return true;
  }

  catch(const char *error_str)
  {
    message_streamt message_stream(get_message_handler());
    message_stream.err_location(symex.last_source_location);
    message_stream.error(error_str);
    return true;
  }

  catch(std::bad_alloc)
  {
    error() << "Out of memory" << eom;
    return true;
  }

  statistics() << "size of program expression: "
               << equation.SSA_steps.size()
               << " steps" << eom;

  try
  {
    if(options.get_option("slice-by-trace")!="")
    {
      symex_slice_by_tracet symex_slice_by_trace(ns);

      symex_slice_by_trace.slice_by_trace
	(options.get_option("slice-by-trace"), equation);
    }

    if(equation.has_threads())
    {
      // we should build a thread-aware SSA slicer
      statistics() << "no slicing due to threads" << eom;
    }
    else
    {
      if(options.get_bool_option("slice-formula"))
      {
        slice(equation);
        statistics() << "slicing removed "
                     << equation.count_ignored_SSA_steps()
                     << " assignments" << eom;
      }
      else
      {
        simple_slice(equation);
        statistics() << "simple slicing removed "
                     << equation.count_ignored_SSA_steps()
                     << " assignments" << eom;
      }
    }

    {
      statistics() << "Generated " << symex.total_claims
                   << " VCC(s), " << symex.remaining_claims
                   << " remaining after simplification" << eom;
    }

    if(options.get_bool_option("show-vcc"))
    {
      show_vcc();
      return false;
    }
    
    if(options.get_bool_option("cover-assertions"))
    {
      satcheckt satcheck;
      satcheck.set_message_handler(get_message_handler());
      bv_cbmct bv_cbmc(ns, satcheck);
      bv_cbmc.set_message_handler(get_message_handler());

      if(options.get_option("arrays-uf")=="never")
        bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
      else if(options.get_option("arrays-uf")=="always")
        bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
    
      cover_assertions(goto_functions, bv_cbmc);
      return false;
    }

    // any properties to check at all?
    if(!options.get_bool_option("program-only") &&
       symex.remaining_claims==0)
    {
      report_success();
      return false;
    }

    if(options.get_bool_option("all-claims"))
    {
      satcheckt satcheck;
      satcheck.set_message_handler(get_message_handler());
      bv_cbmct bv_cbmc(ns, satcheck);
      bv_cbmc.set_message_handler(get_message_handler());

      if(options.get_option("arrays-uf")=="never")
        bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
      else if(options.get_option("arrays-uf")=="always")
        bv_cbmc.unbounded_array=bv_cbmct::U_ALL;

      return all_claims(goto_functions, bv_cbmc);
    }
    
  //  repairmodet rmode = static_cast<repairmodet>(options.get_int_option("repair"));
	if (rm==SATOMIC)
		return sa_repair();
	else if(rm==WATOMIC)
	{//protect guards are empty is an indication that satomic
	  //phase for watomic repair is not over
	  if(satomic_repair.protect_guards.empty())
	    return sa_repair();
	  else
	    return wa_repair();
	}
	else if(rm==PSOTSO ||rm==PSOSC || rm==TSOSC)
	  return wmm_repair();
  /*	else if(options.get_bool_option("boolector"))
          return decide_boolector();
    else if(options.get_bool_option("mathsat"))
      return decide_mathsat();
    else if(options.get_bool_option("cvc"))
      return decide_cvc();*/
	else if(options.get_bool_option("smt1"))
      return decide_smt1();
    else if(options.get_bool_option("smt2"))
      return decide_smt2();
    else if(options.get_bool_option("dimacs"))
      return write_dimacs();
    else if(options.get_bool_option("refine"))
      return decide_bv_refinement();
    else if(options.get_bool_option("aig"))
      return decide_aig();
    else
    {
      if(options.get_bool_option("program-only"))
      {
        show_program();
        return false;
      }

      return decide_default();
    }
  }

  catch(std::string &error_str)
  {
    error(error_str);
    return true;
  }

  catch(const char *error_str)
  {
    error(error_str);
    return true;
  }

  catch(std::bad_alloc)
  {
    error() << "Out of memory" << eom;
    return true;
  }
}

/*******************************************************************\

Function: bmct::decide

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool bmct::decide(prop_convt &prop_conv)
{
  if(options.get_bool_option("beautify-pbs") ||
     options.get_bool_option("beautify-greedy"))
    throw "sorry, this solver does not support beautification";

  prop_conv.set_message_handler(get_message_handler());
  
  bool result=true;

  switch(run_decision_procedure(prop_conv))
  {
  case decision_proceduret::D_UNSATISFIABLE:
    result=false;
    report_success();
    break;

  case decision_proceduret::D_SATISFIABLE:
    error_trace(prop_conv);
    report_failure();
    break;

  default:
    error() << "decision procedure failed" << eom;
  }

  return result;
}

/*******************************************************************\

Function: bmct::setup_unwind

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::setup_unwind()
{
  const std::string &set=options.get_option("unwindset");
  std::string::size_type length=set.length();

  for(std::string::size_type idx=0; idx<length; idx++)
  {
    std::string::size_type next=set.find(",", idx);
    std::string val=set.substr(idx, next-idx);

    unsigned thread_nr;
    bool thread_nr_set=false;

    if(!val.empty() &&
       isdigit(val[0]) &&
       val.find(":")!=std::string::npos)
    {
      std::string nr=val.substr(0, val.find(":"));
      thread_nr=unsafe_string2unsigned(nr);
      thread_nr_set=true;
      val.erase(0, nr.size()+1);
    }

    if(val.rfind(":")!=std::string::npos)
    {
      std::string id=val.substr(0, val.rfind(":"));
      long uw=unsafe_string2int(val.substr(val.rfind(":")+1));

      if(thread_nr_set)
        symex.set_unwind_thread_loop_limit(thread_nr, id, uw);
      else
        symex.set_unwind_loop_limit(id, uw);
    }
    
    if(next==std::string::npos) break;
    idx=next;
  }

  if(options.get_option("unwind")!="")
    symex.set_unwind_limit(options.get_unsigned_int_option("unwind"));
}

void bmct::report_reordering_fix(const std::set<unsigned int>& groups,
		literal2symbolt& literal2symbol,orderingliteralst& oliterals)
{

    for(typename std::set<unsigned int>::const_iterator sit=groups.begin();
    		sit!=groups.end();sit++)
    {

    	const irep_idt sym = literal2symbol[oliterals[*sit].front()];
    	peventt p = symbol2order[sym];

    	std::cout << "Fence between" << p.first->source.pc->source_location.as_string() << " and " \
    			<< p.second->source.pc->source_location.as_string() << " is required" << std::endl;

    }


}
