/*******************************************************************\

Module: Solvers for VCs Generated by Symbolic Execution of ANSI-C

Author: Daniel Kroening, kroening@kroening.com

\*******************************************************************/

#include <memory>
#include <iostream>

#include <solvers/sat/satcheck.h>
#include <solvers/sat/satcheck_minisat2.h>
#include <solvers/sat/satcheck_lingeling.h>

#include <solvers/refinement/bv_refinement.h>
#include <solvers/smt1/smt1_dec.h>
#include <solvers/smt2/smt2_dec.h>
#include <solvers/cvc/cvc_dec.h>

#include <solvers/prop/aig_prop.h>

#include "bmc.h"
#include "bv_cbmc.h"
#include "counterexample_beautification.h"
#include "version.h"
//added by sbjoshi
#include <goto-symex/build_goto_trace.h>
/*******************************************************************\

Function: bmct::get_smt1_solver_type

  Inputs: None

 Outputs: An smt1_dect::solvert giving the solver to use.

 Purpose: Uses the options to pick an SMT 1.2 solver

\*******************************************************************/

smt1_dect::solvert bmct::get_smt1_solver_type() const
{
  assert(options.get_bool_option("smt1"));

  smt1_dect::solvert s = smt1_dect::GENERIC;
  
  if(options.get_bool_option("boolector"))
    s = smt1_dect::BOOLECTOR;
  else if(options.get_bool_option("mathsat"))
    s = smt1_dect::MATHSAT;
  else if(options.get_bool_option("cv3"))
    s = smt1_dect::CVC3;
  else if(options.get_bool_option("cv4"))
    s = smt1_dect::CVC4;
  else if(options.get_bool_option("opensmt"))
    s = smt1_dect::OPENSMT;
  else if(options.get_bool_option("yices"))
    s = smt1_dect::YICES;
  else if(options.get_bool_option("z3"))
    s = smt1_dect::Z3;
  else if(options.get_bool_option("generic"))
    s = smt1_dect::GENERIC;

  return s;
}

/*******************************************************************\

Function: bmct::get_smt2_solver_type

  Inputs: None

 Outputs: An smt2_dect::solvert giving the solver to use.

 Purpose: Uses the options to pick an SMT 2.0 solver

\*******************************************************************/

smt2_dect::solvert bmct::get_smt2_solver_type() const
{
  assert(options.get_bool_option("smt2"));

  smt2_dect::solvert s = smt2_dect::GENERIC;
  
  if(options.get_bool_option("boolector"))
    s = smt2_dect::BOOLECTOR;
  else if(options.get_bool_option("mathsat"))
    s = smt2_dect::MATHSAT;
  else if(options.get_bool_option("cv3"))
    s = smt2_dect::CVC3;
  else if(options.get_bool_option("cv4"))
    s = smt2_dect::CVC4;
  else if(options.get_bool_option("opensmt"))
    s = smt2_dect::OPENSMT;
  else if(options.get_bool_option("yices"))
    s = smt2_dect::YICES;
  else if(options.get_bool_option("z3"))
    s = smt2_dect::Z3;
  else if(options.get_bool_option("generic"))
    s = smt2_dect::GENERIC;

  return s;
}

/*******************************************************************\

Function: bmct::solver_factory

  Inputs:

 Outputs:

 Purpose: Decide using "default" decision procedure

\*******************************************************************/

prop_convt *bmct::solver_factory()
{
  //const std::string &filename=options.get_option("outfile");
  
  if(options.get_bool_option("boolector"))
  {
  }
  else if(options.get_bool_option("mathsat"))
  {
  }
  else if(options.get_bool_option("cvc"))
  {
  }
  else if(options.get_bool_option("dimacs"))
  {
  }
  else if(options.get_bool_option("opensmt"))
  {
  }
  else if(options.get_bool_option("refine"))
  {
  }
  else if(options.get_bool_option("aig"))
  {
  }
  else if(options.get_bool_option("smt1"))
  {
  }
  else if(options.get_bool_option("smt2"))
  {
  }
  else if(options.get_bool_option("yices"))
  {
  }
  else if(options.get_bool_option("z3"))
  {
  }
  else
  {
    // THE DEFAULT

    #if 0
    // SAT preprocessor won't work with beautification.
    if(options.get_bool_option("sat-preprocessor") &&
       !options.get_bool_option("beautify"))
    {
      solver=std::auto_ptr<propt>(new satcheckt);
    }
    else
      solver=std::auto_ptr<propt>(new satcheck_minisat_no_simplifiert);

    solver->set_message_handler(get_message_handler());
      
    bv_cbmct bv_cbmc(ns, *solver);
      
    if(options.get_option("arrays-uf")=="never")
      bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
    else if(options.get_option("arrays-uf")=="always")
      bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
    #endif
  }      

  return 0;
}

/*******************************************************************\

Function: bmct::decide_default

  Inputs:

 Outputs:

 Purpose: Decide using "default" decision procedure

\*******************************************************************/

bool bmct::decide_default()
{
  bool result=true;
  
  std::auto_ptr<propt> solver;

  // SAT preprocessor won't work with beautification.
  if(options.get_bool_option("sat-preprocessor") &&
     !options.get_bool_option("beautify"))
  {
    solver=std::auto_ptr<propt>(new satcheckt);
  }
  else
    solver=std::auto_ptr<propt>(new satcheck_minisat_no_simplifiert);

  solver->set_message_handler(get_message_handler());
    
  bv_cbmct bv_cbmc(ns, *solver);
    
  if(options.get_option("arrays-uf")=="never")
    bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
  else if(options.get_option("arrays-uf")=="always")
    bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
    
  switch(run_decision_procedure(bv_cbmc))
  {
  case decision_proceduret::D_UNSATISFIABLE:
    result=false;
    report_success();
    break;

  case decision_proceduret::D_SATISFIABLE:
    if(options.get_bool_option("beautify"))
      counterexample_beautificationt()(
        bv_cbmc, equation, ns);

    error_trace(bv_cbmc);
    report_failure();
    break;

  default:
    error() << "decision procedure failed" << eom;
  }

  return result;
}

/*******************************************************************\

Function: bmct::decide_aig

  Inputs:

 Outputs:

 Purpose: Decide using AIG followed by SAT

\*******************************************************************/

bool bmct::decide_aig()
{
  bool result=true;

  std::auto_ptr<propt> sub_solver;

  if(options.get_bool_option("sat-preprocessor"))
    sub_solver=std::auto_ptr<propt>(new satcheckt);
  else
    sub_solver=std::auto_ptr<propt>(new satcheck_minisat_no_simplifiert);

  aig_prop_solvert solver(*sub_solver);

  solver.set_message_handler(get_message_handler());
    
  bv_cbmct bv_cbmc(ns, solver);
    
  if(options.get_option("arrays-uf")=="never")
    bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
  else if(options.get_option("arrays-uf")=="always")
    bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
    
  switch(run_decision_procedure(bv_cbmc))
  {
  case decision_proceduret::D_UNSATISFIABLE:
    result=false;
    report_success();
    break;

  case decision_proceduret::D_SATISFIABLE:
    error_trace(bv_cbmc);
    report_failure();
    break;

  default:
    error() << "decision procedure failed" << eom;
  }

  return result;
}

/*******************************************************************\

Function: bmct::bv_refinement

  Inputs:

 Outputs:

 Purpose: Decide using refinement decision procedure

\*******************************************************************/

bool bmct::decide_bv_refinement()
{
  std::auto_ptr<propt> solver;

  // We offer the option to disable the SAT preprocessor
  if(options.get_bool_option("sat-preprocessor"))
    solver=std::auto_ptr<propt>(new satcheckt);
  else
    solver=std::auto_ptr<propt>(new satcheck_minisat_no_simplifiert);
  
  solver->set_message_handler(get_message_handler());

  bv_refinementt bv_refinement(ns, *solver);

  // we allow setting some parameters  
  if(options.get_option("max-node-refinement")!="")
    bv_refinement.max_node_refinement=
      options.get_unsigned_int_option("max-node-refinement");
  
  return decide(bv_refinement);
}

void print_trace(const goto_tracet& goto_trace)
{
	for(goto_tracet::stepst::const_iterator sit=goto_trace.steps.begin();
			sit!=goto_trace.steps.end();sit++)
	{
		std::cout << sit->thread_nr << ":" << sit->pc->function << ":" << sit->pc->location_number << ":" \
				<< sit->pc->source_location.as_string() << std::endl;
	}
}
/** added by sbjoshi **/
bool bmct::wmm_repair()
{
  bool result;
  result=true;

   // simplifier won't work with beautification
  //  satcheck_minisat_no_simplifiert satcheck;
   satcheck_minisat_simplifiert satcheck;
 // satcheck_lingelingt satcheck;
    satcheck.set_message_handler(get_message_handler());
  //  satcheck.set_verbosity(get_verbosity());

    bv_cbmct bv_cbmc(ns, satcheck);

    if(options.get_option("arrays-uf")=="never")
      bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
    else if(options.get_option("arrays-uf")=="always")
      bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
  /***************/
  status("Passing problem to "+bv_cbmc.decision_procedure_text());

  bv_cbmc.set_message_handler(get_message_handler());
  //bv_cbmc.set_verbosity(get_verbosity());

  // stop the time
 // fine_timet sat_start=current_time();

  orderingliteralst oliterals;
  symbol2grpt symbol2grp;
  literal2symbolt literal2symbol;

  sym2grpsymt sym2grpsym;
  grp2grplitt grp2grplit;
  group_from_ordering(symbol2grp,sym2grpsym);
  do_conversion(bv_cbmc);


  populate_oliterals_map(bv_cbmc,literal2symbol,symbol2grp,oliterals,grp2grplit);
#ifdef DEBUG
  std::cout << "after populating literals" << std::endl;
#endif

  bvt elements;
 // unsigned size=oliterals.size();
  mhsit m;
   std::set<unsigned int> sol;
   unsigned int counter = 0;
    bvt prev_assumptions;

    fine_timet formula_build_time_end = current_time();
    time_periodt fm_build_time(formula_build_time_end.get_t()-formula_build_time_start.get_t());
    time_periodt totaltime_for_sat_rounds;
if(options.get_bool_option("reorder-opt"))
{
#ifdef DEBUG
	std::cout << "Searching solution with REORDER OPTIMIZATION" << std::endl;
#endif

		elements.insert(elements.begin(),grp2grplit.begin(),grp2grplit.end());

  /*for(prop_conv_solvert::symbolst::const_iterator cit=bv_cbmc.get_symbols().begin();
		  cit!=bv_cbmc.get_symbols().end();cit++)
  {
	  if(does_irep_id_contains_str((cit->first),GLUE_GROUP_SYMBOL))
		{
			if(!cit->second.is_constant())
			{
				elements.push_back(cit->second);

			}
		}
  }*/
  bv_cbmc.set_frozen(elements);
#ifdef DEBUG
  std::cout << "Size of cardinality literals is : "<< elements.size() << std::endl;
#endif


  unsigned int total_elems=elements.size();
  unsigned int k=options.get_unsigned_int_option("minbound");
  unsigned int k2=options.get_unsigned_int_option("maxbound");
  unsigned int elem_size;
  if(k2!=0 && k2>k)
  {
	  // select the smaller of K2 and the total pairs
	  elem_size=k2>total_elems?total_elems:k2;
  }
  else
  {
	  elem_size=total_elems;
  }
#ifdef DEBUG
  std::cout <<"Minimum bound for ROBMC is " << k << std::endl;
#endif

  if(k>=elem_size)
  {
	  std::cout << "--minbound too high. --minbound should not exceed " << elem_size-1 << "for this program. Aborting..." << std::endl;
	  assert(false);
  }
  status("Running "+bv_cbmc.decision_procedure_text());

  fine_timet sat_rounds_start=current_time();
#ifdef TOTALIZER
  totalizer_encoding_iet te(satcheck,elements,k);
#endif
  while(k<=elem_size)
  {
#ifdef DEBUG
	  std::cout <<"ROBMC bound is now :" << k << std::endl;
#endif
	  bool enc_result;
#ifndef TOTALIZER
  //sequential_encodingt se(satcheck);
	  totalizer_encodingt se(satcheck);
#else
	  prev_assumptions.clear();
	  enc_result=te.atmostk(k,prev_assumptions);
	  assert(enc_result); //encoding must return true
#endif


/*  for(unsigned i=0;i<size;i++)
  {
	elements.insert(elements.begin(),oliterals[i].begin(),oliterals[i].end());
  }*/
#ifndef TOTALIZER
  formulat formula;
  enc_result=se.atmostk(elements,k,formula);
  assert(enc_result);
  literalt cardinality_lit = se.get_enabling_lit_for_formula(formula);
//  satcheck.set_frozen(cardinality_lit); //assumptions are always frozen I think


if(prev_assumptions.empty())
{
  prev_assumptions.push_back(cardinality_lit);
}
else
{
	literalt not_l = !prev_assumptions.back();
	prev_assumptions.pop_back();
	prev_assumptions.push_back(not_l);
	prev_assumptions.push_back(cardinality_lit);
}


#endif



m.get_mhs(sol);
add_reordering_assumptions(bv_cbmc,sol,grp2grplit,prev_assumptions);

  bool repaired=false;

  while(!repaired)
  {
	  unsigned int tmpk;
    counter++;
    std::cout << "Trace : " << counter << std::endl;
     decision_proceduret::resultt dec_result=bv_cbmc.dec_solve();
      goto_tracet goto_trace;
  switch(dec_result)
    {
    case decision_proceduret::D_UNSATISFIABLE:
      result=false;
    repaired=true;
    tmpk=k<<2; //double the k and see if it stays UNSAT
    k=(k==elem_size)?(elem_size+1):(std::min(tmpk,elem_size)); // this assignment to k should be before
    //early-termination check. Otherwise, it might cause infinite loop

    if(options.get_bool_option("early-term"))
    {
#ifndef TOTALIZER
    if(!bv_cbmc.is_in_conflict(cardinality_lit))
#else
    	if(!is_in_conflict(bv_cbmc,prev_assumptions))
#endif
    {
    	std::cout << "early termination" << std::endl;
       	k=elem_size+1;
    }
    }



    /*std::cout << "Program repaired" << std::endl;

    //m.get_mhs(sol);
    {
    satcheck_minisat_no_simplifiert msolver;
    m.get_mhs_from_solver(sol,msolver);
    }
       report_reordering_fix(sol,literal2symbol,oliterals);
      //report_success();*/
      break;

    case decision_proceduret::D_SATISFIABLE:
     // counterexample_beautificationt()(
       // bv_cbmc, equation, ns);
    	build_goto_trace(equation, bv_cbmc, ns, goto_trace);
#ifdef DEBUG
    	print_trace(goto_trace);
    	goto_trace.output(ns,std::cout);
#endif
   update_ordering_constraints(bv_cbmc,m,literal2symbol,symbol2grp,grp2grplit,prev_assumptions);

   //   report_failure();
      break;

    default:
      error("decision procedure failed");
    }


  }
}
  fine_timet sat_rounds_end=current_time();
  totaltime_for_sat_rounds += time_periodt(sat_rounds_end.get_t()-sat_rounds_start.get_t());
}
else
{
	bool repaired=false;

fine_timet sat_rounds_start=current_time();
	  while(!repaired)
	  {
	    counter++;
	    std::cout << "Trace : " << counter << std::endl;
	     decision_proceduret::resultt dec_result=bv_cbmc.dec_solve();
	      goto_tracet goto_trace;
	  switch(dec_result)
	    {
	    case decision_proceduret::D_UNSATISFIABLE:
	      result=false;
	    repaired=true;



	    /*std::cout << "Program repaired" << std::endl;

	    //m.get_mhs(sol);
	    {
	    satcheck_minisat_no_simplifiert msolver;
	    m.get_mhs_from_solver(sol,msolver);
	    }
	       report_reordering_fix(sol,literal2symbol,oliterals);
	      //report_success();*/
	      break;

	    case decision_proceduret::D_SATISFIABLE:
	     // counterexample_beautificationt()(
	       // bv_cbmc, equation, ns);
	    	build_goto_trace(equation, bv_cbmc, ns, goto_trace);
#ifdef DEBUG
	    	print_trace(goto_trace);
	    	goto_trace.output(ns,std::cout);
#endif
	    	if(options.get_bool_option("trace-enum"))
	    	{
	    		update_ordering_constraints(bv_cbmc,satcheck,m,literal2symbol,symbol2grp,grp2grplit);
	    	}
	    	else
	    	{
	   update_ordering_constraints(bv_cbmc,m,literal2symbol,symbol2grp,grp2grplit,prev_assumptions);
	    	}

	   //   report_failure();
	      break;

	    default:
	      error("decision procedure failed");
	    }


	  }
	  fine_timet sat_rounds_end = current_time();
	  totaltime_for_sat_rounds+=time_periodt(sat_rounds_end.get_t()-sat_rounds_start.get_t());

}
 /*
  * REMOVING TEMPORARILY
  * {
      satcheck_minisat_no_simplifiert msolver;
      m.get_mhs_from_solver(sol,msolver);
      }*/
	   m.get_mhs(sol);
         report_reordering_fix(sol,literal2symbol,oliterals);
         std::cout << "Total queries to the solver : " << counter << std::endl;
         std::cout << "Time to build initial formula : " << fm_build_time << std::endl;
         std::cout << "Time for rounds of queries to the solver : " << totaltime_for_sat_rounds << std::endl;
         std::cout << "Time per round is : " << ((double)totaltime_for_sat_rounds.get_t())/(double)(counter*1000) << std::endl;

  return result;
}


/*******************************************************************\

Function: bmct::sa_repair

  Inputs:

 Outputs:

 Purpose: Repair the program by synthesizing strong atomic regions

\*******************************************************************/

bool bmct::sa_repair()

{
  bool result=true;

   // simplifier won't work with beautification
    satcheck_minisat_no_simplifiert satcheck;
    satcheck.set_message_handler(get_message_handler());
    //satcheck.set_verbosity(get_verbosity());
    
    bv_cbmct bv_cbmc(ns, satcheck);
    
    if(options.get_option("arrays-uf")=="never")
      bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
    else if(options.get_option("arrays-uf")=="always")
      bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
	/***************/
	status("Passing problem to "+bv_cbmc.decision_procedure_text());

  bv_cbmc.set_message_handler(get_message_handler());
  //bv_cbmc.set_verbosity(get_verbosity());

  // stop the time
 // fine_timet sat_start=current_time();
  
  do_conversion(bv_cbmc);  
  

  populate_guard_literal_map(bv_cbmc,STRONG);



 mhsit m;
 std::set<unsigned int> sol;
  status("Running "+bv_cbmc.decision_procedure_text());
  unsigned int counter = 0;
  bool repaired=false;
  while(!repaired)
  {
	  counter++;
	  std::cout << "Trace : " << counter << std::endl;
	   decision_proceduret::resultt dec_result=bv_cbmc.dec_solve();
	    goto_tracet goto_trace;
	switch(dec_result)
    {
    case decision_proceduret::D_UNSATISFIABLE:
      result=false;
	  repaired=true;
	  std::cout << "Guards to be protected: ";

	  m.get_mhs(sol);

	  satomic_repair.construct_regions(sol);

	  std::cout << "Program repaired" << std::endl;
      //report_success();
      break;

    case decision_proceduret::D_SATISFIABLE:
     // counterexample_beautificationt()(
       // bv_cbmc, equation, ns);
    	m.get_mhs(sol);
    	if(!sol.empty())
    	{
     print_cs_evaluation(bv_cbmc,sol);
    	}
	 update_atomicity_constraints(equation,bv_cbmc,ns,goto_trace,m,STRONG);
   //   report_failure();
      break;

    default:
      error("decision procedure failed");
    }
 
 
  }


 
  // output runtime

#if 0
  {
    std::ostringstream str;
    fine_timet sat_stop=current_time();

    str << "Runtime decision procedure: ";
    output_time(sat_stop-sat_start, str);
    str << "s";
    status(str.str());
  }

  return dec_result;
  /***************/
#endif
      
  

  return result;
}

void bmct::print_cs_evaluation(const prop_conv_solvert& prop_conv,const std::set<unsigned int>& s)
  {
  	for(std::set<unsigned int>::const_iterator sit=s.begin();sit!=s.end();sit++)
  	{
  		unsigned int guard=(*sit)-1;
  		std::cout<<"Guard " << (guard+1) << "is : " << std::endl;
  	    for(literalst::iterator lit=guard_to_literal_map[guard].begin();
  	    		lit!=guard_to_literal_map[guard].end();lit++)
  	    {
  	        std::cout<< (*lit).var_no() << ":" << prop_conv.l_get(*lit);
  	    }
  	}
  }

bool bmct::wa_repair()

{
  bool result=true;

   // simplifier won't work with beautification
    satcheck_minisat_no_simplifiert satcheck;
    satcheck.set_message_handler(get_message_handler());
   // satcheck.set_verbosity(get_verbosity());

    bv_cbmct bv_cbmc(ns, satcheck);

    if(options.get_option("arrays-uf")=="never")
      bv_cbmc.unbounded_array=bv_cbmct::U_NONE;
    else if(options.get_option("arrays-uf")=="always")
      bv_cbmc.unbounded_array=bv_cbmct::U_ALL;
  /***************/
  status("Passing problem to "+bv_cbmc.decision_procedure_text());

  bv_cbmc.set_message_handler(get_message_handler());
  //bv_cbmc.set_verbosity(get_verbosity());

  // stop the time
 // fine_timet sat_start=current_time();

  do_conversion(bv_cbmc);


  populate_guard_literal_map(bv_cbmc,WEAK);


 mhsit m;

for(typename cpoints_collectiont::iterator it=watomic_repair.cpoints_collection.begin();
     it!=watomic_repair.cpoints_collection.end();it++)
 {
   m.add_to_collection(*it);
   std::cout << "w collection";
#ifdef DEBUG
   std::cout << std::endl;
   std::copy(it->begin(),it->end(),std::ostream_iterator<unsigned int>(std::cout," "));
#endif
 }

std::set<unsigned int> s;
m.get_mhs(s);
std::cout << "wmhs : " ;
std::copy(s.begin(),s.end(),std::ostream_iterator<unsigned int>(std::cout," "));
assert(!s.empty() && "Hitting set of size 0 encountered!!");

add_atomicity_assumptions(bv_cbmc,s,WEAK);
  status("Running "+bv_cbmc.decision_procedure_text());
  unsigned int counter = 0;
  bool repaired=false;
  while(!repaired)
  {
    counter++;
    std::cout << "Trace : " << counter << std::endl;
     decision_proceduret::resultt dec_result=bv_cbmc.dec_solve();
      goto_tracet goto_trace;
  switch(dec_result)
    {
    case decision_proceduret::D_UNSATISFIABLE:
      result=false;
    repaired=true;
    std::cout << "Guards to be protected: ";

    std::cout << "Program repaired" << std::endl;
      //report_success();
      break;

    case decision_proceduret::D_SATISFIABLE:
     // counterexample_beautificationt()(
       // bv_cbmc, equation, ns);

   update_atomicity_constraints(equation,bv_cbmc,ns,goto_trace,m,WEAK);
   //   report_failure();
      break;

    default:
      error("decision procedure failed");
    }


  }


  // output runtime

#if 0
  {
    std::ostringstream str;
    fine_timet sat_stop=current_time();

    str << "Runtime decision procedure: ";
    output_time(sat_stop-sat_start, str);
    str << "s";
    status(str.str());
  }

  return dec_result;
  /***************/
#endif



  return result;
}


/*******************************************************************\

Function: bmct::decide_smt1

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool bmct::decide_smt1()
{
  smt1_dect::solvert solver=get_smt1_solver_type();
  const std::string &filename=options.get_option("outfile");
  
  if(filename=="")
  {
    smt1_dect smt1_dec(
      ns,
      "cbmc",
      "Generated by CBMC " CBMC_VERSION,
      "QF_AUFBV",
      solver);

    return decide(smt1_dec);
  }
  else if(filename=="-")
    smt1_convert(solver, std::cout);
  else
  {
    std::ofstream out(filename.c_str());
    if(!out)
    {
      std::cerr << "failed to open " << filename << std::endl;
      return false;
    }
    
    smt1_convert(solver, out);
  }
  
  return false;
}

/*******************************************************************\

Function: bmct::smt1_convert

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::smt1_convert(smt1_dect::solvert solver, std::ostream &out)
{
  smt1_convt smt1_conv(
    ns,
    "cbmc",
    "Generated by CBMC " CBMC_VERSION,
    "QF_AUFBV",
    solver,
    out);

  smt1_conv.set_message_handler(get_message_handler());
  
  do_conversion(smt1_conv);

  smt1_conv.dec_solve();
}

/*******************************************************************\

Function: bmct::decide_smt2

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

bool bmct::decide_smt2()
{
  smt2_dect::solvert solver=get_smt2_solver_type();
  const std::string &filename=options.get_option("outfile");
  
  if(filename=="")
  {
    smt2_dect smt2_dec(
      ns,
      "cbmc",
      "Generated by CBMC " CBMC_VERSION,
      "QF_AUFBV",
      solver);

    if(options.get_bool_option("fpa"))
      smt2_dec.use_FPA_theory=true;

    return decide(smt2_dec);
  }
  else if(filename=="-")
    smt2_convert(solver, std::cout);
  else
  {
    std::ofstream out(filename.c_str());
    if(!out)
    {
      std::cerr << "failed to open " << filename << std::endl;
      return false;
    }
    
    smt2_convert(solver, out);
  }
  
  return false;
}

/*******************************************************************\

Function: bmct::smt2_convert

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/

void bmct::smt2_convert(
  smt2_dect::solvert solver,
  std::ostream &out)
{
  smt2_convt smt2_conv(
    ns,
    "cbmc",
    "Generated by CBMC " CBMC_VERSION,
    "QF_AUFBV",
    solver,
    out);

  if(options.get_bool_option("fpa"))
    smt2_conv.use_FPA_theory=true;

  smt2_conv.set_message_handler(get_message_handler());
  
  do_conversion(smt2_conv);

  smt2_conv.dec_solve();
}

