#ifndef CPROVER_GLUE_TRANSFORMATION_H
#define CPROVER_GLUE_TRANSFORMATION_H

#include <map>

#include <sstream>
#include <list>
#include "../goto-programs/goto_functions.h"
#include "../goto-programs/goto_program.h"
#include "util/symbol_table.h"
#include "util/symbol.h"
#include "util/expr.h"
#include "util/type.h"
#include "util/irep_ids.h"
#include "util/irep.h"
#include "util/expr_util.h"
#include "util/std_types.h"
#include "util/std_code.h"
#include "util/std_expr.h"
#include "util/arith_tools.h"
#include "util/location.h"
#include "util/hash_cont.h"
#include "glue_util.h"
#include "../goto-programs/remove_skip.h"
#include "../analyses/cfg_dominators.h"


#define Contains(x,y)   ((x).find(y)!=(x).end())

#define GLUE_TCNT  "$glue_thread_counter"
#define GLUE_TCNT_NAME "c::$glue_thread_counter"
#define GLUE_TIDBNAME "$glue_thread_id"
#define GLUE_TIDNAME "c::$glue_thread_id"

#define GLUE_LOCKBNAME "$glue_lock"
#define GLUE_LOCKNAME "c::$glue_lock"
#define GLUE_OLD_LOCKBNAME "$glue_old_lock"
#define GLUE_OLD_LOCKNAME "c::$glue_old_lock"
#define WTMPVAR "___$tmpvar"

// This bitwidth defines how many threads will be allowed. Number of threads allowed will be
// 2^THREAD_BITWIDTH
#define THREAD_BITWIDTH 10


class typet;
class guardt;
class symbol_exprt;
class exprt;

typedef std::list<std::pair<unsigned int,unsigned int> > guard_tracet;
typedef std::set<unsigned int> cpointst;
typedef std::list<cpointst> cpoints_collectiont;
     typedef std::list<guard_tracet> guard_tracest;

//namespace glue{
  /*************
   class transformation_passt : A parent class for program transformation
  **************/
   class transformation_passt
   {
   protected:
    
     /// Input context
       symbol_tablet context_in;
     /// Output context
     symbol_tablet context_out;
      /// Input goto_function
       goto_functionst func_in;
     /// Output goto function (after the transformation)
       goto_functionst func_out;
     
       bool get_tmp_bool_symbol(symbolt& symbol, const irep_idt& mname);
     

   public:
       transformation_passt(const goto_functionst& _in, const symbol_tablet& _ctxt_in):context_in(_ctxt_in),context_out(_ctxt_in)
       {
       
       
       
       
       func_in.copy_from(_in);
       func_out.copy_from(func_in);
       

      
       }

     transformation_passt()
     {}
     virtual void set_params(const goto_functionst& _in, const symbol_tablet& _ctxt_in)
     {
       context_in = _ctxt_in;
       context_out = _ctxt_in;
       func_in.copy_from(_in);
       func_out.copy_from(func_in);


     }

     const goto_functionst & get_input_func()
     {
       return func_in;
     }

     goto_functionst & get_output_func()
     {
       return func_out;
     }

     const symbol_tablet & get_in_context()
     {
       return context_in;
     }

      symbol_tablet & get_out_context()
     {
       return context_out;
     }
       virtual void apply_transformation() = 0;



   } ;

  /* class satomic_transformationt : public transformation_passt
   {

   private:
     /// counter for boolean guards that surrounds the context switch
     unsigned int bguard_counter;

     
     void get_guarded_atomic_gadget(goto_programt& result);
        public:
     satomic_transformationt(const goto_functionst& _in, const contextt & _ctxt_in):transformation_passt(_in, _ctxt_in)
     {
       bguard_counter=0;
     }
     void apply_transformation();
     
   };*/

   class satomic_transformationt:public transformation_passt 
   {
     unsigned int bguard_counter;
     symbolt thread_counter;
     symbolt thread_identifier;
     symbolt glue_lock;
     symbolt glue_old_lock;

     goto_functionst func_weak;


     typedef std::pair<irep_idt,goto_programt::targett> fname_to_instr_pairt;

         typedef std::map<unsigned int,fname_to_instr_pairt> guard_to_instr_mapt;
         guard_to_instr_mapt guard_to_instr_map;

#define LOCKED true
#define UNLOCKED false
    typedef bool lockstatust;
    typedef std::map<goto_programt::const_targett,lockstatust> instr_lockstatus_mapt;
    instr_lockstatus_mapt instr_lockstatus_map;

    typedef std::map<unsigned int , goto_programt::const_targetst > guard_lock_mapt;
    typedef std::map<unsigned int , goto_programt::targetst > guard_unlock_mapt;

    guard_lock_mapt guard_lock_map;
    guard_unlock_mapt guard_unlock_map;

   // typedef std::map<irep_idt,cfg_dominatorst> program_to_cfg_mapt;
   // program_to_cfg_mapt program_to_cfg_map;



    //void compute_dominator(const irep_idt& fname);

    void init_func_weak();
    void points_to_regions();
    void remove_aux_vars();
   public:
    typedef std::set<unsigned int> protect_guardst;

   private:
    void populate_guards_to_protect(const protect_guardst& s);

    bool point_to_region(std::set<unsigned int>& guards_done,
        const unsigned int guard_no, goto_programt& prog,
        const goto_programt::targett& local_lock);


    void build_guard_to_instr_map();


    

     void get_guarded_lock_gadget(goto_programt & result,goto_programt::targett & instr_loc, bool mod_orig=true);
     

     void get_lock_acquire_gadget(goto_programt& result,const irep_idt& fname);
     

     
     void get_lock_release_gadget(goto_programt & result,const irep_idt& fname);
     



     void get_thread_id_gadget(goto_programt& result);
     
  //   bool is_builtin_instruciton(const goto_programt::targett& instr) const;

     void insert_thread_id();
     void remove_auxiliary();
     void insert_locks();
     

   public : 


         protect_guardst protect_guards;

         cpoints_collectiont cpoints_collection;

    // guard_tracest guard_traces;
     void print_regions(std::ostream& out) const;
     void construct_regions(const protect_guardst& s);

     goto_functionst& get_weak_func()
     {
       return func_weak;
     }
     satomic_transformationt()
     {}

     unsigned int get_bguard_counter()
     {
       return bguard_counter;
     }

     void set_params(const goto_functionst& _in, const symbol_tablet& _ctxt_in)
     {
       transformation_passt::set_params(_in,_ctxt_in);
        bguard_counter=0;

       //add thread_counter and thread_identifier to simulate thread identifier
       thread_counter.clear();
       thread_counter.is_static_lifetime=true;
       thread_counter.is_thread_local=false;
       const irep_idt tmp_bname = GLUE_TCNT;
       const irep_idt tmp_name = GLUE_TCNT_NAME;
       thread_counter.mode = ID_C;
       thread_counter.name = tmp_name;
       thread_counter.base_name = tmp_bname;
       thread_counter.type = unsignedbv_typet(THREAD_BITWIDTH);
       thread_counter.value = from_integer(0,thread_counter.type);
       context_out.add(thread_counter);

       thread_identifier.clear();
       thread_identifier.is_static_lifetime=true;
       thread_identifier.is_thread_local = true;
       const irep_idt tmp_tidbname = GLUE_TIDBNAME;
       const irep_idt tmp_tidname = GLUE_TIDNAME;
       thread_identifier.base_name = tmp_tidbname;
       thread_identifier.name = tmp_tidname;
       thread_identifier.mode = ID_C;
       thread_identifier.type = unsignedbv_typet(THREAD_BITWIDTH);
       context_out.add(thread_identifier);


       glue_lock.clear();
       glue_lock.is_static_lifetime=true;
       glue_lock.is_thread_local=false;
       const irep_idt tmp_lockbname = GLUE_LOCKBNAME;
       const irep_idt tmp_lockname = GLUE_LOCKNAME;
       glue_lock.base_name = tmp_lockbname;
       glue_lock.name = tmp_lockname;
       glue_lock.mode = ID_C;
       glue_lock.type = unsignedbv_typet(THREAD_BITWIDTH);
       glue_lock.value = from_integer(0,glue_lock.type);
       context_out.add(glue_lock);

       glue_old_lock.clear();
       glue_old_lock.is_static_lifetime=true;
       glue_old_lock.is_thread_local=false;
       const irep_idt tmp_olockbname = GLUE_OLD_LOCKBNAME;
       const irep_idt tmp_olockname = GLUE_OLD_LOCKNAME;
       glue_old_lock.base_name = tmp_olockbname;
       glue_old_lock.name = tmp_olockname;
       glue_old_lock.mode = ID_C;
       glue_old_lock.type = unsignedbv_typet(THREAD_BITWIDTH);
       glue_old_lock.value = from_integer(0,glue_old_lock.type);
       context_out.add(glue_old_lock);

     }
     satomic_transformationt(const goto_functionst& _in, const symbol_tablet & _ctxt_in):transformation_passt(_in, _ctxt_in)
     {
       bguard_counter=0;

       //add thread_counter and thread_identifier to simulate thread identifier
       thread_counter.clear();
       thread_counter.is_static_lifetime=true;
       thread_counter.is_thread_local=false;
       const irep_idt tmp_bname = GLUE_TCNT;
       const irep_idt tmp_name = GLUE_TCNT_NAME;
       thread_counter.mode = ID_C;
       thread_counter.name = tmp_name;
       thread_counter.base_name = tmp_bname;
       thread_counter.type = unsignedbv_typet(THREAD_BITWIDTH);
       thread_counter.value = from_integer(0,thread_counter.type);
       context_out.add(thread_counter);

       thread_identifier.clear();
       thread_identifier.is_static_lifetime=true;
       thread_identifier.is_thread_local = true;
       const irep_idt tmp_tidbname = GLUE_TIDBNAME;
       const irep_idt tmp_tidname = GLUE_TIDNAME;
       thread_identifier.base_name = tmp_tidbname;
       thread_identifier.name = tmp_tidname;
       thread_identifier.mode = ID_C;
       thread_identifier.type = unsignedbv_typet(THREAD_BITWIDTH);
       context_out.add(thread_identifier);


       glue_lock.clear();
       glue_lock.is_static_lifetime=true;
       glue_lock.is_thread_local=false;
       const irep_idt tmp_lockbname = GLUE_LOCKBNAME;
       const irep_idt tmp_lockname = GLUE_LOCKNAME;
       glue_lock.base_name = tmp_lockbname;
       glue_lock.name = tmp_lockname;
       glue_lock.mode = ID_C;
       glue_lock.type = unsignedbv_typet(THREAD_BITWIDTH);
       glue_lock.value = from_integer(0,glue_lock.type);
       context_out.add(glue_lock);

       glue_old_lock.clear();
       glue_old_lock.is_static_lifetime=true;
       glue_old_lock.is_thread_local=false;
       const irep_idt tmp_olockbname = GLUE_OLD_LOCKBNAME;
       const irep_idt tmp_olockname = GLUE_OLD_LOCKNAME;
       glue_old_lock.base_name = tmp_olockbname;
       glue_old_lock.name = tmp_olockname;
       glue_old_lock.mode = ID_C;
       glue_old_lock.type = unsignedbv_typet(THREAD_BITWIDTH);
       glue_old_lock.value = from_integer(0,glue_old_lock.type);
       context_out.add(glue_old_lock);


     }
     
     void apply_transformation();
   };

   

   class watomic_transformationt : public transformation_passt
   {
     unsigned int wbguard_counter;

      
       symbolt thread_counter;
    symbolt thread_identifier;
     symbolt glue_lock;
    symbolt glue_old_lock;



   /* typedef std::list<goto_programt::targett> unlock_listt;

    unlock_listt unlock_list;
    typedef std::list<goto_programt::targett> lock_listt;
    lock_listt lock_list;*/


     void get_assume_unlocked_gadget(goto_programt & result, goto_programt::targett & instr_loc);


     void simulate_guard_trace(const uisett& guard_trace,
         const std::set<unsigned int>& csguards,
         cpointst& lprotect);

   public :

     cpoints_collectiont cpoints_collection;

     void print_guard_trace(const guard_tracet& guard_trace,std::ostream& out);
     void simulate_guard_traces(
              const cpoints_collectiont& guard_traces,
              const std::set<unsigned int>& csguards);

     typedef std::map<unsigned int,unsigned int> guard_mapt;
     guard_mapt guard_map;
     watomic_transformationt()
     {}
     unsigned int get_wbguard_counter() { return wbguard_counter;}

     void set_params(const goto_functionst& _in, const symbol_tablet& _ctxt_in)
     {
        transformation_passt::set_params(_in,_ctxt_in);
       wbguard_counter = 0;


         thread_counter.clear();
       thread_counter.is_static_lifetime=true;
       thread_counter.is_thread_local=false;
       const irep_idt tmp_bname = GLUE_TCNT;
       const irep_idt tmp_name = GLUE_TCNT_NAME;
       thread_counter.mode = ID_C;
       thread_counter.name = tmp_name;
       thread_counter.base_name = tmp_bname;
       thread_counter.type = unsignedbv_typet(THREAD_BITWIDTH);
       thread_counter.value = from_integer(0,thread_counter.type);
       context_out.add(thread_counter);

       thread_identifier.clear();
       thread_identifier.is_static_lifetime=true;
       thread_identifier.is_thread_local = true;
       const irep_idt tmp_tidbname = GLUE_TIDBNAME;
       const irep_idt tmp_tidname = GLUE_TIDNAME;
       thread_identifier.base_name = tmp_tidbname;
       thread_identifier.name = tmp_tidname;
       thread_identifier.mode = ID_C;
       thread_identifier.type = unsignedbv_typet(THREAD_BITWIDTH);
       context_out.add(thread_identifier);



        glue_lock.clear();
       glue_lock.is_static_lifetime=true;
       glue_lock.is_thread_local=false;
       const irep_idt tmp_lockbname = GLUE_LOCKBNAME;
       const irep_idt tmp_lockname = GLUE_LOCKNAME;
       glue_lock.base_name = tmp_lockbname;
       glue_lock.name = tmp_lockname;
       glue_lock.mode = ID_C;
       glue_lock.type = unsignedbv_typet(THREAD_BITWIDTH);
       glue_lock.value = from_integer(0,glue_lock.type);
       context_out.add(glue_lock);

        glue_old_lock.clear();
       glue_old_lock.is_static_lifetime=true;
       glue_old_lock.is_thread_local=false;
       const irep_idt tmp_olockbname = GLUE_OLD_LOCKBNAME;
       const irep_idt tmp_olockname = GLUE_OLD_LOCKNAME;
       glue_old_lock.base_name = tmp_olockbname;
       glue_old_lock.name = tmp_olockname;
       glue_old_lock.mode = ID_C;
       glue_old_lock.type = unsignedbv_typet(THREAD_BITWIDTH);
       glue_old_lock.value = from_integer(0,glue_old_lock.type);
       context_out.add(glue_old_lock);
     }

     void apply_transformation();

   };

#endif
