#include "glue_util.h"

/************************************************************************\

Function : has_sharedvar_access

Inputs : expr , context

Output : true/false

Purpose : returns whether a given expression accesses a shared variable, if the expression has 
          operands, then recurse on each of them to find out if the symbol is shared or not in the
		  given context

\************************************************************************/
bool has_sharedvar_access(const exprt &expr,const symbol_tablet& context)
{
	bool result = false;
	if(expr.id()==ID_symbol && !expr.has_operands())
	{
		symbol_exprt sym_expr = to_symbol_expr(expr);

		const irep_idt &identifier = sym_expr.get_identifier();
		//contextt::symbolst::const_iterator sym_it = context.symbols.find(identifier);
		if(context.has_symbol(identifier))
		{
			return context.lookup(identifier).is_shared();
		}
		/*if(sym_it != context.symbols.end())
		{
			const symbolt& sym = sym_it->second;
			return sym.is_shared();
		}*/

		
		
	}
	else if(expr.id()!=ID_nil && expr.has_operands())
	{
		forall_operands(it,expr)
		{
			result = result || has_sharedvar_access((*it),context);
		}
	}
	return result;
}


void make_auxiliary(goto_programt::targett& instr)
{
  prepend_to_instruction(instr,GLUEAUX);
}

void make_original(goto_programt::targett& instr)
{
   prepend_to_instruction(instr,GLUEORIG);

}

bool does_instr_contain(const goto_programt::const_targett& instr, const std::string str)
{
   const irep_idt comm = instr->source_location.get_comment();
   if(comm.empty())
   {
     return false;
   }

   std::string s = as_string(comm);
   std::string::size_type found = s.find(str);
   return found==std::string::npos?false:true;
}

void prepend_to_instruction(goto_programt::targett& instr,const std::string str)
{
  const irep_idt comm = instr->source_location.get_comment();

  if(comm.empty())
  {
    instr->source_location.set_comment(str);
    return;
  }

  std::string s(str+" "+as_string(comm));
  instr->source_location.set_comment(s);

}

/****************************************************************************\
 * Function : is_builtin_instruction
 *
 * Input : goto_programt::targett iterator "instr" to the instruction
 *
 * Output : true if the instruction pointed to by instr is an instruction
 * inside a builtin function, false otherwise
 *
 * Purpose :
 ****************************************************************************/


	   bool is_builtin_instruction(const goto_programt::targett& instr)
	   {

		   std::string str1("<built");
		   if(does_irep_id_contains_str(instr->function,str1)) return true;


		   std::string::size_type found =  instr->source_location.as_string().find(str1);

		   if(found!=std::string::npos)
		   {
		     return true;
		   }
		   return false;

	   }
/*****************************************************************************\
 * Function : get_guard_from_instruction
 *
 * Input : goto_programt::targett& instr - iterator to the instruction
 *
 * Output : The number of the guard associated with the instruction, zero
 *              otherwise
 *
 * Precondition : Assume that the comment of the instruction will be non-empty
 *  if it was inserted for the guard. Also the format of the comment would be
 *
 *  XYZ GLUE_BGUARDNUM
 *
 *  where GLUE_BGUARD is the macro-expansion value and NUM is the number and the
 *  comment ends here
 */

unsigned int get_guard_from_instruction(const goto_programt::const_targett& instr,atomicityt a)
{

	const irep_idt comm=instr->source_location.get_comment();

	if(comm.empty())
		return 0;
	std::string s = as_string(comm);

	std::string bguardstr = (a==STRONG)?GLUE_BGUARD:GLUE_WBGUARD;
	 std::size_t bguardstrlen = bguardstr.length();
	std::string::size_type found = s.find(bguardstr);

	if(found==std::string::npos)
	{
		return 0;
	}
	else
	{
		std::string ss = s.substr(found+bguardstrlen);
		std::istringstream is(ss);
		unsigned int guardno;
		is >> guardno;

		if((is.rdstate() & std::istringstream::failbit)
				||( is.rdstate() & (std::istringstream::badbit)) )
		   {
		       return 0;
		   }
		   return guardno;

	}

	return 0;


}


/*****************************************************************************\
 * Function : is_auxiliary_instruction
 *
 * Input : goto_programt::targett instr - iterator to the instruction
 *
 * Output : if the instruction was the auxiliary instruction added during the
 * transformation
 *
 * Precondition : During the transformation apply_transformation() the
 *  instructions
 * which are inserted will have "Auxiliary" in their comment
 */
bool is_auxiliary_instruction(const goto_programt::targett& instr)
{

	const irep_idt comm= instr->source_location.get_comment();
	if(comm.empty()) return false;

   std::string s = as_string(comm);

   std::string::size_type found = s.find(GLUEAUX);

   if(found!=std::string::npos)
   {
	   return true;
   }
   return false;
}
bool is_auxiliary_instruction(const goto_programt::const_targett& instr)
{

	const irep_idt comm= instr->source_location.get_comment();
	if(comm.empty()) return false;

   std::string s = as_string(comm);

   std::string::size_type found = s.find(GLUEAUX);

   if(found!=std::string::npos)
   {
	   return true;
   }
   return false;
}

bool is_original_instruction(const goto_programt::const_targett& instr)
{

  const irep_idt comm= instr->source_location.get_comment();
  if(comm.empty()) return false;

   std::string s = as_string(comm);

   std::string::size_type found = s.find(GLUEORIG);

   if(found!=std::string::npos)
   {
     return true;
   }
   return false;
}

unsigned int get_guard_from_trace_step(const goto_trace_stept& e,atomicityt a)
{
        unsigned int cstaken=0;

        /**DO NOT search for guard in builtin functions/libraries. At this point of time,
         * it is not clear
        whether we should leave the guards inside these regions outside the search or not.
        As of now, keeping it outside **/
        std::string filename = e.pc->source_location.as_string();
      //  std::cout << filename << std::endl;
        std::string str1("<builtin");
        std::string::size_type found  = filename.find(str1);
        std::string bguardstr;
        if(a==STRONG)
        {
        bguardstr = GLUE_BGUARD;
        }
        else
        {
         bguardstr = GLUE_WBGUARD;
        }
        std::size_t bguardstrlen = bguardstr.length();


        if(found!=std::string::npos)
        {
        //        std::cout << "found at " << found <<" "<< std::string::npos << std::endl;
                return 0;
        }

        // see if the comment is of the form "GLUEBGUARD : num" where num is the number of the guard
        const irep_idt e_comm = e.pc->source_location.get_comment();
        if(!e_comm.empty())
        {
                std::string s = as_string(e_comm);
                found = s.find(bguardstr);
                if(found==std::string::npos)
                {
                	return 0;
                }


                     assert(s.length() > bguardstrlen);
                     std::string guard_no = s.substr(found+bguardstrlen);
                         std::istringstream ss(guard_no);
                         ss >> cstaken;
                         if((ss.rdstate() & std::istringstream::failbit)
                        		 || (ss.rdstate() & (std::istringstream::badbit)))
                         {
                                 return 0;
                         }
                         return cstaken;



        }
        return 0;

}
/****************************************************************************\
 * Function : get_conflict_points_from_trace
 *
 * Input : goto_trace
 *
 * Output : cpoints
 *
 * Purpose : It is assumed that after
 *  watomic_transformationt::apply_transformation whenever a lock is acquired
 *  or released
 *  the instruction comment will have GLUELOCKING or GLUEUNLOCKING respectively.
 *  Furthermore, whenever assume unlocked gadget is encountered the
 *  comment will have GLUEUNLOCKED.
 *
 *  Therefore, in the trace, you wait till the lock is acquired, then
 *  all the shared var access is considered conflict points and then you add them
 *  in cpoints. Keep doing it until you hit a lock release.
 *
 */

void get_conflict_points_from_trace(const goto_tracet&goto_trace,
    std::set<unsigned int>& cpoints)
{
  bool islocked=false;
  for(goto_tracet::stepst::const_iterator s_it=goto_trace.steps.begin();
      s_it!=goto_trace.steps.end();s_it++)
  {
#ifdef DEBUG
//sbjoshi
	  std::cout << s_it->pc->function << ":" << s_it->pc->location_number << ":" << s_it->pc->source_location.get_comment() << std::endl;
#endif
      const irep_idt c = s_it->pc->source_location.get_comment();
      if(c.empty()) continue;



      if(!islocked && does_irep_id_contains_str(c,GLUELOCKING))
      {
    	  std::cout << "acquiring lock" << std::endl;
        islocked=true;
      }
      else if(islocked && does_irep_id_contains_str(c,GLUEUNLOCKING))
      {
    	  std::cout << "releasing lock" << std::endl;
        islocked=false;
      }
      else if(islocked && does_irep_id_contains_str(c,GLUEUNLOCKED))
      {
        unsigned int g = get_guard_from_instruction(s_it->pc,WEAK);
        if(g!=0)
        {
        	std::cout << "cpoint inside lock" << g << std::endl;
          cpoints.insert(g);
        }
      }
  }
}

void extract_conflict_points(const guard_tracet& guard_trace,
		std::set<unsigned int>& cpoints,const unsigned int thread_nr,
		unsigned int cs)
{
	for(guard_tracet::const_reverse_iterator git=guard_trace.rbegin();
			git!=guard_trace.rend();
			git++)
	{
		if(git->second==0) continue;
		if(git->second!=cs) cpoints.insert(git->second);
		if(git->second==cs && git->first==thread_nr) break;
	}

}
/*****************************************************************************\
 * Function : get_cs_taken_from_trace
 *
 * Input : goto_trace, cs_taken, guard_trace
 *
 * Output :
 *
 * Purpose : Given a goto_trace (of tyep goto_tracet) extract the context
 * switches taken by the trace and store it in the set cs_taken. In addition,
 * store the guard_trace. guard_trace is the consists of pairs of thread_nr
 * and guard_no (from satomic_transformationt). However, when the context switch
 * occurs, we want to store source and target thread_nr irrespective
 *  if the corresponding
 * instructions had a valid guard. This guard_trace will be used for WATOMIC
 * repair. Once we form lock regions from strong atomicity, we will use these
 * traces to emulate the queries and see which shared access (guards) were in
 * conflict with strong atomicity regions. This way, guard_trace helps us
 * avoid firing some of the actual queries to SAT solver that would have been
 * needed otherwise
 */
void get_cstaken_from_trace(const goto_tracet& goto_trace,
    std::set<unsigned int>& cs_taken,
    std::set<unsigned int>& cpoints)
{

	std::list<std::pair<unsigned int,unsigned int> > guard_trace;
	typedef std::map<unsigned int,unsigned int> tr2guardt;
	tr2guardt tr2guard;
	 goto_tracet::stepst::const_iterator s_it1=goto_trace.steps.begin();
	 goto_tracet::stepst::const_iterator s_end=goto_trace.steps.end();
	 unsigned int cstaken=0,lastorigthread=0;
	//	 unsigned prev_cs=0,prev_thread=0;
		 while(s_it1!=s_end)
		 {
			 goto_tracet::stepst::const_iterator next=(++s_it1);
			 s_it1--;

			 std::cout << s_it1->thread_nr << " : "<< s_it1->pc->function << ":" << s_it1->pc->source_location << " :" << s_it1->pc->location_number << " : " \
					 << s_it1->pc->source_location.get_comment() << std::endl;
			 if(next!=s_end)
			 {
			 std::cout << next->thread_nr << " : " << next->pc->function << ":" << next->pc->source_location << " :" << next->pc->location_number << " : " \
			 					 << next->pc->source_location.get_comment() << std::endl;
			 }


			bool is_original = is_original_instruction(s_it1->pc);
			 if(is_original)
			 {
			   cstaken = get_guard_from_trace_step((*s_it1),STRONG);
			   assert(cstaken!=0);

			   if(lastorigthread!=s_it1->thread_nr)
			   {
				   tr2guardt::iterator it=tr2guard.find(s_it1->thread_nr);
				  				     if(it!=tr2guard.end() && it->second!=0)
				  				     {
				  				    	 extract_conflict_points(guard_trace,cpoints,
				  				    			 s_it1->thread_nr,it->second);
				  				    	 cs_taken.insert(it->second);
				  				    	 std::cout << "cs inserted is : " << it->second << std::endl;
				  				     }

			   }
		/*	   if(prev_thread!=s_it1->thread_nr)
			   {
				   if(prev_cs!=0)
				   {
				   cs_taken.insert(prev_cs);
				   std::cout << "cs inserted is : " << prev_cs << std::endl;
				   }
			   }
			   prev_thread=s_it1->thread_nr;
			   prev_cs=cstaken;*/
			   lastorigthread=s_it1->thread_nr;
			   tr2guard[lastorigthread]=cstaken;
			   guard_trace.push_back(pair_uint(s_it1->thread_nr,cstaken));
			 }
			 if(is_auxiliary_instruction(s_it1->pc)
					 && does_instr_contain(s_it1->pc,GLUEUNLOCKING)
					 && !does_instr_contain(s_it1->pc,GLUE_BGUARD))
			 {
				 //unconditional unlocking does not qualify for context switch
				 tr2guardt::iterator it=tr2guard.find(s_it1->thread_nr);
				 if(it!=tr2guard.end())
				 {
					 it->second=0;
				 }
			 }
			 if(next!=s_end && s_it1->thread_nr!=next->thread_nr)
			 {
				 std::cout << "thread changed" << std::endl;
				 /*if(next->thread_nr!=lastorigthread)
				 {
				     tr2guardt::iterator it=tr2guard.find(next->thread_nr);
				     if(it!=tr2guard.end())
				     {
				    	 cs_taken.insert(it->second);
				    	 std::cout << "cs inserted is : " << it->second << std::endl;
				     }
				 }*/
				//unsigned int cstaken =  get_guard_from_trace_step((*s_it1),STRONG);
			//	if(does_instr_contain(s_it1->pc,GLUE_BGUARD))
			//	{
				//	cstaken=get_guard_from_instruction(s_it1->pc,STRONG);

				//	assert(cstaken!=0);


				/*	if(cstaken!=0)
					{
						std::cout << "cs inserted is : " << cstaken << std::endl;
					cs_taken.insert(cstaken);

					}*/
			//	}
			//	else
			//	{
				  //store the thread_nr from where the context switch is taken in the trace
				  guard_trace.push_back(pair_uint(s_it1->thread_nr,0));
			//	}
				  //if next->pc is "ORIGINAL" instruction then it will be considered in the next iteration
				  // store the target thread_nr of the context switch
				  if(!is_original_instruction(next->pc))
				  {
				    guard_trace.push_back(pair_uint(next->thread_nr,0));
				  }



			 }
		 s_it1++;


		 }
//when a context switch happens from any thread t1, without t1 being unconditionally unlocked
// and never returns to t1 and the last thread accessing shared var is not t1 then include those
// context switches --
		 for(tr2guardt::const_iterator tit=tr2guard.begin();
				 tit!=tr2guard.end();tit++)
		 {
			 if(tit->first!=lastorigthread && tit->second!=0)
			 {
				cs_taken.insert(tit->second);
				extract_conflict_points(guard_trace,cpoints,tit->first,tit->second);
			 }

		 }





}

#if 0
void get_cstaken_from_trace(const goto_tracet& goto_trace,
    std::set<unsigned int>& cs_taken,
    std::set<unsigned int>& cpoints)
{

	std::list<std::pair<unsigned int,unsigned int> > guard_trace;
	typedef std::map<unsigned int,unsigned int> tr2guardt;
	tr2guardt tr2guard;
	 goto_tracet::stepst::const_iterator s_it1=goto_trace.steps.begin();
	 goto_tracet::stepst::const_iterator s_end=goto_trace.steps.end();
	 unsigned int cstaken=0,lastorigthread=0;
	//	 unsigned prev_cs=0,prev_thread=0;
		 while(s_it1!=s_end)
		 {
			 goto_tracet::stepst::const_iterator next=(++s_it1);
			 s_it1--;

			 std::cout << s_it1->thread_nr << " : "<< s_it1->pc->function << ":" << s_it1->pc->location << " :" << s_it1->pc->location_number << " : " \
					 << s_it1->pc->location.get_comment() << std::endl;
			 if(next!=s_end)
			 {
			 std::cout << next->thread_nr << " : " << next->pc->function << ":" << next->pc->location << " :" << next->pc->location_number << " : " \
			 					 << next->pc->location.get_comment() << std::endl;
			 }


			bool is_original = is_original_instruction(s_it1->pc);
			 if(is_original)
			 {
			   cstaken = get_guard_from_trace_step((*s_it1),STRONG);
			   assert(cstaken!=0);

			   if(lastorigthread!=s_it1->thread_nr)
			   {
				   tr2guardt::iterator it=tr2guard.find(s_it1->thread_nr);
				  				     if(it!=tr2guard.end() && it->second!=0)
				  				     {
				  				    	 extract_conflict_points(guard_trace,cpoints,
				  				    			 s_it1->thread_nr,it->second);
				  				    	 cs_taken.insert(it->second);
				  				    	 std::cout << "cs inserted is : " << it->second << std::endl;
				  				     }

			   }
		/*	   if(prev_thread!=s_it1->thread_nr)
			   {
				   if(prev_cs!=0)
				   {
				   cs_taken.insert(prev_cs);
				   std::cout << "cs inserted is : " << prev_cs << std::endl;
				   }
			   }
			   prev_thread=s_it1->thread_nr;
			   prev_cs=cstaken;*/
			   lastorigthread=s_it1->thread_nr;
			   tr2guard[lastorigthread]=cstaken;
			   guard_trace.push_back(pair_uint(s_it1->thread_nr,cstaken));
			 }
			 if(is_auxiliary_instruction(s_it1->pc)
					 && does_instr_contain(s_it1->pc,GLUEUNLOCKING)
					 && !does_instr_contain(s_it1->pc,GLUE_BGUARD))
			 {
				 //unconditional unlocking does not qualify for context switch
				 tr2guardt::iterator it=tr2guard.find(s_it1->thread_nr);
				 if(it!=tr2guard.end())
				 {
					 it->second=0;
				 }
			 }
			 if(next!=s_end && s_it1->thread_nr!=next->thread_nr)
			 {
				 std::cout << "thread changed" << std::endl;
				 /*if(next->thread_nr!=lastorigthread)
				 {
				     tr2guardt::iterator it=tr2guard.find(next->thread_nr);
				     if(it!=tr2guard.end())
				     {
				    	 cs_taken.insert(it->second);
				    	 std::cout << "cs inserted is : " << it->second << std::endl;
				     }
				 }*/
				//unsigned int cstaken =  get_guard_from_trace_step((*s_it1),STRONG);
			//	if(does_instr_contain(s_it1->pc,GLUE_BGUARD))
			//	{
				//	cstaken=get_guard_from_instruction(s_it1->pc,STRONG);

				//	assert(cstaken!=0);


				/*	if(cstaken!=0)
					{
						std::cout << "cs inserted is : " << cstaken << std::endl;
					cs_taken.insert(cstaken);

					}*/
			//	}
			//	else
			//	{
				  //store the thread_nr from where the context switch is taken in the trace
				  guard_trace.push_back(pair_uint(s_it1->thread_nr,0));
			//	}
				  //if next->pc is "ORIGINAL" instruction then it will be considered in the next iteration
				  // store the target thread_nr of the context switch
				  if(!is_original_instruction(next->pc))
				  {
				    guard_trace.push_back(pair_uint(next->thread_nr,0));
				  }



			 }
		 s_it1++;


		 }





}

#endif

void get_conflictpoints_from_trace(const goto_tracet& goto_trace,std::set<unsigned int>& cp)
{
  unsigned int lockthread;
  bool flag=false;

  for(goto_tracet::stepst::const_iterator s_it=goto_trace.steps.begin();
      s_it!=goto_trace.steps.end();s_it++)
  {
     if(does_instr_contain(s_it->pc,GLUELOCKING))
     {
        lockthread=s_it->thread_nr;
        flag=true;

     }

     else if(flag && does_instr_contain(s_it->pc,GLUELOCKED))
     {
       unsigned int g = get_guard_from_trace_step((*s_it),WEAK);
       if(g!=0)
       {
         cp.insert(g);
       }
     }

     else if(does_instr_contain(s_it->pc,GLUEUNLOCKING) && s_it->thread_nr==lockthread)
     {
       flag = false;
     }
  }
}

bool does_irep_id_contains_str(const irep_idt& id,const std::string& s)
{
  std::string str=as_string(id);
  std::string::size_type found = str.find(s);
  return found==std::string::npos?false:true;
}

bool does_expr_uses_symbol_containing(const exprt& expr,const std::string& s)
{
  bool result = false;
  if(expr.id()==ID_symbol && !expr.has_operands())
  {
    symbol_exprt sym_expr = to_symbol_expr(expr);

    const irep_idt &identifier = sym_expr.get_identifier();

    return does_irep_id_contains_str(identifier,s);



  }
  else if(expr.id()!=ID_nil && expr.has_operands())
  {
    forall_operands(it,expr)
    {
      result = result || does_expr_uses_symbol_containing(*it,s);
    }
  }
  return result;

}

bool is_pthread_method(const goto_programt::targett& instr)
{

   if(!instr->is_function_call()) return false;
    const irep_idt fname = to_symbol_expr(to_code_function_call(instr->code).function()).get_identifier();
    bool result= does_irep_id_contains_str(fname,"pthread");
    return result|| does_irep_id_contains_str(fname,"__CPROVER");
}

bool is_built_in_event(const partial_order_concurrencyt::eventt &e1,const partial_order_concurrencyt::eventt &e2)
{
	std::string s("<built-in-additions>");
	std::string s1("__CPROVER_initialize");
	bool result = does_irep_id_contains_str(e1.source.pc->source_location.get_file(),s);
	if(result) return true;
	result = does_irep_id_contains_str(e1.source.pc->function,s1);
	if(result) return true;
	result = does_irep_id_contains_str(e2.source.pc->function,s1);
		if(result) return true;

	return does_irep_id_contains_str(e2.source.pc->source_location.get_file(),s);
}

bool is_in_conflict(const prop_convt& prop, const bvt& assumptions)
{
	unsigned size = assumptions.size();
	bool result=false;

	for(unsigned int i=0;i<size && !result;i++)
	{
	     	result = result || prop.is_in_conflict(assumptions[i]);
	}
return result;

}

