/*******************************************************************\

Module: Memory model for partial order concurrency

Author: Michael Tautschnig, michael.tautschnig@cs.ox.ac.uk

\*******************************************************************/

#include <util/std_expr.h>
#include <util/simplify_expr.h>

#include "memory_model_pso_repair.h"

#ifdef DEBUG
#include <iostream>
#endif


/*******************************************************************\

Function: memory_model_tsot::program_order

  Inputs:

 Outputs:

 Purpose:

\*******************************************************************/
#if 0
void memory_model_pso_repairt::program_order(
  symex_target_equationt &equation)
{
  per_thread_mapt per_thread_map;
  build_per_thread_map(equation, per_thread_map);

  thread_spawn(equation, per_thread_map);
  
  // iterate over threads

  for(per_thread_mapt::const_iterator
      t_it=per_thread_map.begin();
      t_it!=per_thread_map.end();
      t_it++)
  {
    const event_listt &events=t_it->second;
    
    // iterate over relevant events in the thread
    
    for(event_listt::const_iterator
        e_it=events.begin();
        e_it!=events.end();
        e_it++)
    {
      if(is_memory_barrier(*e_it))
        continue;

      event_listt::const_iterator next=e_it;
      ++next;

      exprt mb_guard_r, mb_guard_w;
      mb_guard_r.make_false();
      mb_guard_w.make_false();

      for(event_listt::const_iterator
          e_it2=next;
          e_it2!=events.end();
          e_it2++)
      {
        if(is_spawn(*e_it) ||
           is_spawn(*e_it2))
        {
          equation.constraint(
            before(*e_it, *e_it2),
            "po",
            (*e_it)->source);

          if(is_spawn(*e_it2))
            break;
          else
            continue;
        }

        if(is_memory_barrier(*e_it2))
        {
          const codet &code=to_code((*e_it)->source.pc->code);

          if(is_shared_read(*e_it) &&
             !code.get_bool(ID_RRfence) &&
             !code.get_bool(ID_RWfence))
            continue;
          else if(is_shared_write(*e_it) &&
             !code.get_bool(ID_WRfence) &&
             !code.get_bool(ID_WWfence))
            continue;

          if(code.get_bool(ID_RRfence) ||
             code.get_bool(ID_WRfence))
            mb_guard_r=or_exprt(mb_guard_r, (*e_it2)->guard);

          if(code.get_bool(ID_RWfence) ||
             code.get_bool(ID_WWfence))
            mb_guard_w=or_exprt(mb_guard_w, (*e_it2)->guard);

          continue;
        }
        bool porelaxed = program_order_is_relaxed(*e_it,*e_it2);

               if(porelaxed)

               {

                 if(repairto==SC || !memory_model_tsot::program_order_is_relaxed(*e_it,*e_it2))
                 {
                 symbol_exprt e_sym = get_ordering_symbol();

                           equation.constraint(
                               implies_exprt(e_sym,before(*e_it,*e_it2)),
                               GLUEREPAIRSTR,(*e_it)->source
                               );
                           equation.constraint( implies_exprt(
                               not_exprt(e_sym),before(*e_it2,*e_it)),
                               GLUEREPAIRSTR,(*e_it)->source
                               );
                           symbol2order[e_sym.get_identifier()] = peventt(*e_it,*e_it2);
                 }
               }


        exprt cond;
        if(porelaxed)
        {

          if(is_shared_read(*e_it2))
            cond=mb_guard_r;
          else
            cond=mb_guard_w;
        }
        else
          cond.make_true();
        simplify(cond, ns);

        if(!cond.is_false())
          equation.constraint(
            implies_exprt(cond, before(*e_it, *e_it2)),
            "po",
            (*e_it)->source);
      }
    }
  }
}

#endif
void memory_model_pso_repairt::program_order(
  symex_target_equationt &equation)
{
  per_thread_mapt per_thread_map;
  build_per_thread_map(equation, per_thread_map);

  thread_spawn(equation, per_thread_map);

  // iterate over threads

  for(per_thread_mapt::const_iterator
      t_it=per_thread_map.begin();
      t_it!=per_thread_map.end();
      t_it++)
  {
    const event_listt &events=t_it->second;

    // iterate over relevant events in the thread

    for(event_listt::const_iterator
        e_it=events.begin();
        e_it!=events.end();
        e_it++)
    {
      if(is_memory_barrier(*e_it))
        continue;

      event_listt::const_iterator next=e_it;
      ++next;

      exprt mb_guard_r, mb_guard_w;
      mb_guard_r.make_false();
      mb_guard_w.make_false();

      std::list<symbol_exprt> prev_symbol;

      for(event_listt::const_iterator
          e_it2=next;
          e_it2!=events.end();
          e_it2++)
      {
        if((is_spawn(*e_it) && !is_memory_barrier(*e_it2)) ||
           is_spawn(*e_it2))
        {
          add_constraint(
            equation,
            before(*e_it, *e_it2),
            "po",
            (*e_it)->source);

          if(is_spawn(*e_it2))
            break;
          else
            continue;
        }

        if(is_memory_barrier(*e_it2))
        {
          const codet &code=to_code((*e_it2)->source.pc->code);

          if(is_shared_read(*e_it) &&
             !code.get_bool(ID_RRfence) &&
             !code.get_bool(ID_RWfence))
            continue;
          else if(is_shared_write(*e_it) &&
             !code.get_bool(ID_WRfence) &&
             !code.get_bool(ID_WWfence))
            continue;

          if(code.get_bool(ID_RRfence) ||
             code.get_bool(ID_WRfence))
            mb_guard_r=or_exprt(mb_guard_r, (*e_it2)->guard);

          if(code.get_bool(ID_RWfence) ||
             code.get_bool(ID_WWfence))
            mb_guard_w=or_exprt(mb_guard_w, (*e_it2)->guard);

          continue;
        }

        exprt cond=true_exprt();
        exprt ordering=nil_exprt();

        bool porelaxed = program_order_is_relaxed(*e_it,*e_it2);
        if(address(*e_it)==address(*e_it2))
        {
          ordering=partial_order_concurrencyt::before(
            *e_it, *e_it2, AX_SC_PER_LOCATION);
        }
        else if(porelaxed)
               {
        	if(repairto==SC || !memory_model_tsot::program_order_is_relaxed(*e_it,*e_it2))
        	{
        		if(!is_built_in_event(*(*e_it),*(*e_it2)))
        		{
                 symbol_exprt e_sym = get_ordering_symbol();

                           equation.constraint(
                               implies_exprt(e_sym,before(*e_it2,*e_it)),
                               GLUEREPAIRSTR,(*e_it)->source
                               );
                           equation.constraint( implies_exprt(
                               not_exprt(e_sym),before(*e_it,*e_it2)),
                               GLUEREPAIRSTR,(*e_it)->source
                               );
                           //wrt e_it, if two events e_it2 and e_it2' can be relaxed than
                           // if e_it2 precedes e_it2', if we want to prevent reordering of
                           //e_it and e_it2, we also should prevent reordering of e_it and e_it2'
                                                     if(!prev_symbol.empty())
                                                     {
                                                  	   symbol_exprt& p_symbol=prev_symbol.back();
                                                  	   equation.constraint( implies_exprt(
                                                  			   not_exprt(p_symbol),not_exprt(e_sym)),
                                                  			   GLUEREPAIRSTR,(*e_it)->source
                                                  			   );

#ifdef DEBUG
                                                  	   std::cout << p_symbol.get_identifier() << " implies" << e_sym.get_identifier() << std::endl;
#endif
                                                  	  prev_symbol.pop_back();

                                                     }
#ifdef DEBUG
                           std::cout << e_sym.get_identifier() << ":" << (*e_it)->source.pc->source_location.as_string() << ":" << \
                        		                              		(*e_it2)->source.pc->source_location.as_string() << std::endl;
#endif
                           symbol2order[e_sym.get_identifier()] = peventt(*e_it,*e_it2);
                           prev_symbol.push_back(e_sym);
        		}

                           if(is_shared_read(*e_it2))
                             cond=mb_guard_r;
                           else
                             cond=mb_guard_w;

                           simplify(cond, ns);
        	}
               }
        if(!cond.is_false())
        {
          if(ordering.is_nil())
            ordering=partial_order_concurrencyt::before(
              *e_it, *e_it2, AX_PROPAGATION);

          add_constraint(
            equation,
            implies_exprt(cond, ordering),
            "po",
            (*e_it)->source);
        }
      }
    }
  }
}
