/*
 * util.h
 *
 *  Created on: 29 Oct 2012
 *      Author: joshi
 */

#ifndef CPROVER_GLUE_UTIL_H_
#define CPROVER_GLUE_UTIL_H_

#include <boost/foreach.hpp>
#include<algorithm>
#include <set>
#include <list>
#include <vector>
#include <string>
#include <map>
#include <iterator>
#include <limits>
#include <iostream>
#include <cassert>
//using namespace std;

#define DECLARE_TYPES(T)	typedef list<set<T> > collectiont;	\
				typedef typename collectiont::iterator collection_itt;	\
				typedef set<T> sett;	\
				typedef typename sett::iterator set_itt;	


#define foreach BOOST_FOREACH
#define reverse_foreach BOOST_REVERSE_FOREACH





namespace glue {





template<typename T,typename U>
	void print_map(const std::map<T,U>& m)
	{
		std::pair<T,U> p;
		foreach(p, m)
		{
			std::cout<< p.first << " : " << p.second << std::endl ;
		}
	}

template<typename T>
	void print_collection(const std::list<std::set<T> >& c)
	{
		foreach(const std::set<T>& s,   c)
		{
			std::copy(s.begin(),s.end(),std::ostream_iterator<T>(std::cout," "));
		}
		std::cout << std::endl;
	}
	
template<typename T>
std::set<T> setUnion(const std::list<std::set<T> >& collection)
{
	std::set<T> result;
	foreach(std::set<T> s,collection)
	{
		foreach(T e, s)
		{
			result.insert(e);
		}
	}

	return result;
}


template<typename T,typename U>
void update_list_map(std::map<T,std::list<U> >& m,T& elem, U& s)
{

	if(m.find(elem)==m.end())
	{
		std::list<U> l;
		l.push_back(s);
		m[elem]=l;

	}
	else
	{
		m[elem].push_back(s);
	}
}

template<typename T>
void update_histogram(std::map<T,unsigned int>& m,T& elem)
{
	if(m.find(elem)==m.end())
	{
		m[elem] = 1;
	}
	else
	{
		m[elem]++;
	}
}

template<typename T>
std::map<T,unsigned int> computeHistogram(const std::list<std::set<T> >& collection)
{
	std::map<T,unsigned int> result;
	foreach(std::set<T> s, collection)
	{
		foreach(T e, s)
		{
			if(result.find(e)==result.end())
			{
				result[e]=1;
			}
			else
			{
				result[e]++;
			}
		}
	}
}


template<typename T>
std::list<T> listOfKeys(const std::map<T,unsigned int> m)
{
	std::list<T> result;
	std::pair<T,unsigned int> p;
	foreach(p, m)
	{
		result.push_back(p.first);
	}
}

template<typename T>
std::set<T> setOfKeys(const std::map<T,unsigned int> m)
{
	std::set<T> result;
	std::pair<T,unsigned int> p;
	foreach( p, m)
	{
		result.insert(p.first);
	}
}

template<typename T>
std::list<T> listOfValues(const std::map<T,unsigned int> m)
{
	std::list<T> result;
	std::pair<T,unsigned int> p;
	foreach(p, m)
	{
		result.push_back(p.second);
	}
}

template<typename T>
std::set<T> setOfValues(const std::map<T,unsigned int> m)
{
	std::set<T> result;
	std::pair<T,unsigned int> p;
	foreach(p, m)
	{
		result.insert(p.second);
	}
}






template<typename T>
std::pair<T,unsigned int> hasGreaterFreq(const std::pair<T,unsigned int> a, const std::pair<T,unsigned int> b)
{
	return a.second > b.second;
}


/*template<typename T>
set<T> set_union(list<set<T> >& collection)
{
	DECLARE_TYPES(T);
	set<T> result;
	for(collection_itt it = collection.begin();it!=collection.end();it++)
	{
		for(set_itt its = it->begin();its!=it->end();its++)
		{
			result.insert((*its));
		}
	}



	return result;

}*/

/*
template<typename T>
list<T> list_merge(list<set<T> >& collection)
{
	list<T> result;

	DECLARE_TYPES(T);

	for(collection_itt it = collection.begin();it!=collection.end();it++)
	{
		for(set_itt its= it->begin();its!= it->end();its++)
		{
			result.push_back((*its));
		}
	}


	return result;
}*/




}



#endif /* UTIL_H_ */
